# coding: utf-8
"""
Utility functions and classes for validation.
"""
__author__ = "Dan Gunter"
__copyright__ = "Copyright 2012-2013, The Materials Project"
__version__ = "1.0"
__maintainer__ = "Dan Gunter"
__email__ = "dkgunter@lbl.gov"
__status__ = "Development"
__date__ = "3/29/13"

#!/usr/bin/env python3


from argparse import Action
from collections import deque
from itertools import chain
import logging
import time
from sys import getsizeof
from yaml import load

TRACE = logging.DEBUG -1


class DoesLogging:
    """Mix-in class that creates the attribute 'log', setting its qualified
    name to the name of the module and class.
    """
    def __init__(self, name=None):
        if name is None:
            if self.__module__ != '__main__':
                name = "%s.%s" % (self.__module__, self.__class__.__name__)
            else:
                name = self.__class__.__name__
        self._log = logging.getLogger(name)
        # cache whether log is debug or higher in a flag to
        # lower overhead of debugging statements
        self._dbg = self._log.isEnabledFor(logging.DEBUG)
        self._trace = self._log.isEnabledFor(TRACE)


def total_size(o, handlers={}, verbose=False, count=False):
    """Returns the approximate memory footprint an object and all of its contents.

    Automatically finds the contents of the following builtin containers and
    their subclasses:  tuple, list, deque, dict, set and frozenset.
    To search other containers, add handlers to iterate over their contents:

        handlers = {SomeContainerClass: iter,
                    OtherContainerClass: OtherContainerClass.get_elements}

    Source: http://code.activestate.com/recipes/577504/ (r3)
    """
    # How to make different types of objects iterable
    dict_handler = lambda d: chain.from_iterable(d.items())
    all_handlers = {tuple: iter,
                    list: iter,
                    deque: iter,
                    dict: dict_handler,
                    set: iter,
                    frozenset: iter}
    all_handlers.update(handlers)     # user handlers take precedence
    seen = set()                      # track which object id's have already been seen
    default_size = getsizeof(0)       # estimate sizeof object without __sizeof__

    def sizeof(o):
        "Calculate size of `o` and all its children"
        if id(o) in seen:             # do not double count the same object
            return 0
        seen.add(id(o))
        if count:
            s = 1
        else:
            s = getsizeof(o, default_size)
        # If `o` is iterable, add size of its members
        for typ, handler in all_handlers.items():
            if isinstance(o, typ):
                s += sum(map(sizeof, handler(o)))
                break
        return s

    return sizeof(o)


class ElapsedTime:
    def __init__(self):
        self.value = -1


class Timing:
    """Perform and report timings using the 'with' keyword.

    For example:
        with Timing('foo', info='bar'):
            do_foo1()
            do_foo2()
    """
    def __init__(self, name="event", elapsed=None, log=None, level=logging.DEBUG, **kwargs):
        self.name, self.kw, self.level = name, kwargs, level
        self.elapsed = elapsed
        self._log = log

    def __enter__(self):
        self.begin = time.time()

    def __exit__(self, type, value, tb):
        elapsed = time.time() - self.begin
        if self._log is not None:
            nvp = ', '.join(['{}={}'.format(k, v) for k, v in self.kw.items()])
            self._log.log(self.level, '@{n}={s:f}s {kw}'.format(n=self.name, s=elapsed, kw=nvp))
        if self.elapsed:
            self.elapsed.value = elapsed


def letter_num(x, letter='A'):
    s, a0 = '', ord(letter) - 1
    while x > 0:
        s = chr(a0 + x % 26) + s
        x /= 26
    return s


class JsonWalker:
    """Walk a dict, transforming.
    Used for JSON formatting.
    """
    def __init__(self, value_transform=None, dict_transform=None):
        """Constructor.

        :param value_transform: Apply this function to each value in a list or dict.
        :type value_transform: function taking a single arg (the value)
        :param dict_transform: Apply this function to each dict
        :type dict_transform: function taking a single arg (the dict)
        """
        self._vx = value_transform
        self._dx = dict_transform

    def walk(self, o):
        """Walk a dict & transform.
        """
        if isinstance(o, dict):
            d = o if self._dx is None else self._dx(o)
            return {k: self.walk(v) for k, v in d.items()}
        elif isinstance(o, list):
            return [self.walk(v) for v in o]
        else:
            return o if self._vx is None else self._vx(o)

    @staticmethod
    def value_json(o):
        """Apply as_json() method on object to get value,
        otherwise return object itself as the value.
        """
        if hasattr(o, 'as_json'):
            return o.as_json()
        return o

    @staticmethod
    def dict_expand(o):
        """Expand keys in a dict with '.' in them into
        sub-dictionaries, e.g.

        {'a.b.c': 'foo'} ==> {'a': {'b': {'c': 'foo'}}}
        """
        r = {}
        for k, v in o.items():
            if isinstance(k, str):
                k = k.replace('$', '_')
            if "." in k:
                sub_r, keys = r, k.split('.')
                # create sub-dicts until last part of key
                for k2 in keys[:-1]:
                    sub_r[k2] = {}
                    sub_r = sub_r[k2]  # descend
                    # assign last part of key to value
                sub_r[keys[-1]] = v
            else:
                r[k] = v
        return r

# Argument handling
# -----------------

_alog = logging.getLogger("mg.args")
#_alog.setLevel(logging.DEBUG)
_argparse_is_dumb = True  # because it doesn't report orig. error text


class YamlConfig(Action):
    """Populate arguments with YAML file contents.

    Adapted from:
      http://code.activestate.com/recipes/577918-filling-command-line-arguments-with-a-file/
    """

    def __call__(self, parser, namespace, values, option_string=None):
        config = self._get_config_from_file(values)
        for key, value in config.items():
            setattr(namespace, key, value)
        _alog.debug("YamlConfig.namespace={}".format(namespace))

    def _get_config_from_file(self, filename):
        with open(filename) as f:
            config = load(f)
        return config


def args_kvp_nodup(s):
    """Parse argument string as key=value pairs separated by commas.

    :param s: Argument string
    :return: Parsed value
    :rtype: dict
    :raises: ValueError for format violations or a duplicated key.
    """
    if s is None:
        return {}
    d = {}
    for item in [e.strip() for e in s.split(",")]:
        try:
            key, value = item.split("=", 1)
        except ValueError:
            msg = "argument item '{}' not in form key=value".format(item)
            if _argparse_is_dumb:
                _alog.warn(msg)
            raise ValueError(msg)
        if key in d:
            msg = "Duplicate key for '{}' not allowed".format(key)
            if _argparse_is_dumb:
                _alog.warn(msg)
            raise ValueError(msg)
        d[key] = value
    return d


def args_list(s):
    """Parse argument string as list of values separated by commas.

    :param s: Argument string
    :return: Parsed value
    :rtype: list
    """
    if s is None:
        return []
    return [item.strip() for item in s.split(',')]



"""
Shared code for builders.

For developers implementing a new builder,
you should inherit from :class:`Builder`.
"""
__author__ = 'Dan Gunter <dkgunter@lbl.gov>'
__date__ = '5/29/13'

## Imports

# system
from abc import ABCMeta, abstractmethod
import copy
import logging
import multiprocessing
try:
    import Queue
except ImportError:
    import queue as Queue
import traceback
# local
from matgendb.builders import schema, util
from matgendb import util as dbutil

## Logging

_log = util.get_builder_log("core")

## Exceptions


class BuildError(Exception):
    def __init__(self, who, why):
        errmsg = "Builder {} failed: {}".format(who, why)
        Exception.__init__(self, errmsg)

## Versioning (experimental)

DB_VERSION = 1

## Functions


def parse_fn_docstring(fn):
    """Get parameter and return types from function's docstring.

    Docstrings must use this format::

       :param foo: What is foo
       :type foo: int
       :return: What is returned
       :rtype: double

    :return: A map of names, each with keys 'type' and 'desc'.
    :rtype: tuple(dict)
    """
    doc = fn.__doc__
    params, return_ = {}, {}
    param_order = []
    for line in doc.split("\n"):
        line = line.strip()
        if line.startswith(":param"):
            _, name, desc = line.split(":", 2)
            name = name[6:].strip()  # skip 'param '
            params[name] = {'desc': desc.strip()}
            param_order.append(name)
        elif line.startswith(":type"):
            _, name, desc = line.split(":", 2)
            name = name[5:].strip()  # skip 'type '
            if not name in params:
                raise ValueError("'type' without 'param' for {}".format(name))
            params[name]['type'] = desc.strip()
        elif line.startswith(":return"):
            _1, _2, desc = line.split(":", 2)
            return_['desc'] = desc
        elif line.startswith(":rtype"):
            _1, _2, desc = line.split(":", 2)
            return_['type'] = desc.strip()
    return params #, return_

## Classes


class Collections:
    """Interface to normalized names for collections.

    After initialization with a MongoDB database and optional parameters,
    you can access collections in `known_collections` as attributes.
    """
    #: Collection names that are accessible as attributes
    #: of an instance of this class
    known_collections = [
        'tasks', 'materials', 'diffraction_patterns',
        'electrodes', 'conversion_electrodes', 'bandstructures', 'icsd',
        'phase_diagrams', 'brototypes', 'electronic_structure']

    MIN_VER = 1
    MAX_VER = 1

    def __init__(self, db, version=1, prefix=None, task_suffix=None):
        """Set collections from database.

        :param db: MongoDB database, but really anything that acts like a dict. If None, it will be ignored.
        :type db: dict-like object
        :param version: Version of naming scheme for the collections
        :type version: int
        :param prefix: Prefix string to put before collection names, e.g. "dahn". Full
                       collection name will be <prefix>.<name>; don't include '.' in the input.
        :type prefix: str
        :param task_suffix: Add this suffix to the tasks collection. Used for merged collections for sandboxes.
        :type task_suffix: str
        :raise: ValueError if `version` is not known
        """
        if not self.MIN_VER <= version <= self.MAX_VER:
            raise ValueError("Bad version ({v:d}) not in range {v0} .. {v1} ".
                             format(v=version, v0=self.MIN_VER, v1=self.MAX_VER))
        self._names, self._coll = {}, {}
        if version == 1:
            for name in self.known_collections:
                full_name = "{}.{}".format(prefix, name) if prefix else name
                if name == 'tasks' and task_suffix is not None:
                    full_name = "{}.{}".format(full_name, task_suffix)
                self._names[name] = full_name
                self._coll[full_name] = None
        if db is None:
            self._db = dict.fromkeys(self.known_collections, 1)
        else:
            self._db = db
        self._prefix = prefix

    def __getattr__(self, item):
        if item in self._names:
            coll_name = self._names[item]
            coll = self._coll[coll_name]
            if coll is None:
                self._coll[coll_name] = coll = self._db[coll_name]
            return coll
        return self.__dict__[item]

    def get_collection_name(self, alias):
        return self._names[alias]

    @property
    def database(self):
        """Return the current database object.

        :return: Current database object
        """
        return self._db


def merge_tasks(core_collections, sandbox_collections, id_prefix, new_tasks, batch_size=100, wipe=False):
    """Merge core and sandbox collections into a temporary collection in the sandbox.

    :param core_collections: Core collection info
    :type core_collections: Collections
    :param sandbox_collections: Sandbox collection info
    :type sandbox_collections: Collections
    """
    merged = copy.copy(sandbox_collections)
    # create/clear target collection
    target = merged.database[new_tasks]
    if wipe:
        _log.debug("merge_tasks.wipe.begin")
        target.remove()
        merged.database['counter'].remove()
        _log.debug("merge_tasks.wipe.end")
    # perform the merge
    batch = []
    for doc in core_collections.tasks.find():
        batch.append(doc)
        if len(batch) == batch_size:
            target.insert(batch)
            batch = []
    if batch:
        target.insert(batch)
    batch = []
    for doc in sandbox_collections.tasks.find():
        doc['task_id'] = id_prefix + '-' + str(doc['task_id'])
        batch.append(doc)
        if len(batch) == batch_size:
            target.insert(batch)
            batch = []
    if batch:
        target.insert(batch)


class HasExamples:
    """Mix-in class for checking the output of a builder.

    This is a way to get some static checking of the schema of inserted documents,
    without imposing the burden of schema-checking every single document.
    The check is static in the sense that it will only be run by the unit tests.

     See the
    `pymatpro.db.builders.test.test_porous_builder` module for an example of
    how to use this to perform unit tests.
    """

    def examples(self):
        """Return example document(s) for collection(s).

        This must be implemented in subclasses.

        :return: List of pairs (doc, collection_name)
        :rtype: list(dict,str)
        """
        raise NotImplementedError()


    def validate_examples(self, fail_fn):
        """Check the examples against the schema.

        :param fail_fn: Pass failure messages to this function
        :type fail_fn: function(str)
        """
        for collection, doc in self.examples():
            _log.debug("validating example in collection {}".format(collection))
            sch = schema.get_schema(collection)  # with more err. checking
            result = sch.validate(doc)
            _log.debug("validation result: {}".format("OK" if result is None else result))
            if result is not None:
                fail_fn("Failed to validate sample document: {}".format(result))


class Builder(metaclass=ABCMeta):
    """Abstract base class for all builders

    To implement a new builder, inherit from this class and
    define the :meth:`get_items` and :meth:`process_item` methods.
    See the online documentation for details.
    """

    def __init__(self, ncores=1):
        """Create new builder for threaded or multiprocess execution.

        :param ncores: Desired number of threads or processes to run
        :type ncores: int
        :raise: ValueError for bad 'config' arg
        """
        sequential = (ncores == 1)
        if sequential:
            self._seq = True
            self._queue = Queue.Queue()
        else:
            self._seq = False
            self._mgr = multiprocessing.Manager()
            self._ncores = ncores if ncores > 0 else 15
            self._queue = multiprocessing.Queue()
            self._run_parallel_fn = self._run_parallel_multiprocess
        self._status = BuilderStatus(ncores, self)

    # ----------------------------
    # Override these in subclasses
    # ----------------------------

    def get_parameters(self):
        """Return key/value pairs that will be passed to get_items().

        This is an alternative to the use of special docstrings as described
        in :meth:`get_items`.

        If the type of the argument is 'QueryEngine', then the driver program
        will add options for, and create, a matgendb.query_engine.QueryEngine
        instance. The value given for this argument will be interpreted as
        the MongoDB collection name.

        Some other basic Python types -- list, dict, int, float -- are
        automatically parsed. If an empty value is supplied for these, then
        None will be passed.

        :return: {'param_name': {'type': 'param_type', 'desc':'description'},
                  'param2_name': {'type': 'param2_type', 'desc':'descr2'}, ..}
        :rtype: dict
        """
        return None  # means: use docstring

    @abstractmethod
    def get_items(self):
        """Perform one-time setup at the top of a run, returning
        an iterator on items to use as input.

        If get_parameters() doesn't return None, this method is used to
        discover the names and types of this function's parameters.
        Otherwise, the docstring of this function is used.
        This docstring must use the ':' version of the restructured
        text style. For example::

            class MyBuilder(Builder):
                def get_items(self, source=None, target=None):
                '''
                :param source: The input porous materials collection
                :type source: QueryEngine
                :param target: The output materials collection
                :type target: QueryEngine
                '''

        More details on the parameters is in :meth:`get_items_parameters`.

        :return: iterator
        """
        return [{"Hello": 1}, {"World": 2}]

    @abstractmethod
    def process_item(self, item):
        """Implement the analysis for each item of work here.

        :param item: One item of work from the queue (i.e., one item from the iterator that
                     was returned by the `setup` method).
        :type item: object
        :return: Status code, 0 for OK
        :rtype: int
        """
        return 0

    def finalize(self, had_errors):
        """Perform any cleanup actions after all items have been processed.
        Subclasses may choose not to implement this, in which case it is a no-op.

        :param had_errors: True if the run itself had errors.
        :type had_errors: bool
        :return: True if nothing went wrong, else False
        """
        return True

    # -----------------------------
    # Utility methods
    # -----------------------------

    def shared_dict(self):
        """Get dict that can be shared between parallel processes.
        """
        if self._seq:
            return dict()
        else:
            return self._mgr.dict()

    def shared_list(self):
        """Get list that can be shared between parallel processes.
        """
        if self._seq:
            return list()
        else:
            return self._mgr.list()

    # -----------------------------
    # Public/internal
    # -----------------------------

    def run(self, user_kw=None, build_kw=None):
        """Run the builder.

        :param user_kw: keywords from user
        :type user_kw: dict
        :param build_kw: internal settings
        :type build_kw: dict
        :return: Number of items processed
        :rtype: int
        """
        user_kw = {} if user_kw is None else user_kw
        build_kw = {} if build_kw is None else build_kw
        n = self._build(self.get_items(**user_kw), **build_kw)
        finalized = self.finalize(self._status.has_failures())
        if not finalized:
            _log.error("Finalization failed")
        return n

    def connect(self, config):
        """Connect to database with given configuration, which may be a dict or
        a path to a pymatgen-db configuration.
        """
        if isinstance(config, str):
            conn = dbutil.get_database(config_file=config)
        elif isinstance(config, dict):
            conn = dbutil.get_database(settings=config)
        else:
            raise ValueError("Configuration, '{}',  must be a path to "
                             "a configuration file or dict".format(config))
        return conn

    # -----------------------------

    def _build(self, items, chunk_size=10000):
        """Build the output, in chunks.

        :return: Number of items processed
        :rtype: int
        """
        _log.debug("_build, chunk_size={:d}".format(chunk_size))
        n, i = 0, 0
        for i, item in enumerate(items):
            if i == 0:
                _log.debug("_build, first item")
            if 0 == (i + 1) % chunk_size:
                if self._seq:
                    self._run(0)
                else:
                    self._run_parallel_fn()  # process the chunk
                if self._status.has_failures():
                    break
                n = i + 1
            self._queue.put(item)
        # process final chunk
        if self._seq:
            self._run(0)
        else:
            self._run_parallel_fn()
        if not self._status.has_failures():
            n = i + 1
        return n

    def _run_parallel_multiprocess(self):
        """Run processes from queue
        """
        _log.debug("run.parallel.multiprocess.start")
        processes = []
        ProcRunner.instance = self
        for i in range(self._ncores):
            self._status.running(i)
            proc = multiprocessing.Process(target=ProcRunner.run, args=(i,))
            proc.start()
            processes.append(proc)
        for i in range(self._ncores):
            processes[i].join()
            code = processes[i].exitcode
            self._status.success(i) if 0 == code else self._status.fail(i)
        _log.debug("run.parallel.multiprocess.end states={}".format(self._status))

    def _run(self, index):
        """Run method for one thread or process
        Just pull an item off the queue and process it,
        until the queue is empty.

        :param index: Sequential index of this process or thread
        :type index: int
        """
        while 1:
            try:
                item = self._queue.get(timeout=2)
                self.process_item(item)
            except Queue.Empty:
                break
            except Exception as err:
                _log.error("In _run(): {}".format(err))
                if _log.isEnabledFor(logging.DEBUG):
                    _log.error(traceback.format_exc())
                self._status.fail(index)
                raise
        self._status.success(index)

    def __str__(self):
        return self.__class__.__name__


class BuilderStatus:
    """Status of a Builder object run.
    """
    # States
    WAIT, RUNNING, SUCCESS, FAILURE = 0, 1, 2, -1
    # For printing.
    _NAMES = {WAIT: "wait", RUNNING: "run", SUCCESS: "ok", FAILURE: "fail"}

    def __init__(self, num, builder):
        self._bld = builder
        self._states = self._bld.shared_list()
        self._states.extend([self.WAIT] * num)

    def running(self, i):
        """Set state of a single process or thread to 'running'.

        :param i: Index of process or thread.
        :return: None
        """
        self._set(i, self.RUNNING)

    def success(self, i):
        """Set state of a single process or thread to 'success'.

        :param i: Index of process or thread.
        :return: None
        """
        self._set(i, self.SUCCESS)

    def fail(self, i):
        """Set state of a single process or thread to 'failure'.

        :param i: Index of process or thread.
        :return: None
        """
        self._set(i, self.FAILURE)

    def has_failures(self):
        """Whether there are any failures in the states.
        """
        return self.FAILURE in self._states

    def _set(self, index, value):
        self._states[index] = value

    def __getitem__(self, index):
        return self._states[index]

    def __str__(self):
        return ",".join([self._NAMES[state] for state in self._states])


class ProcRunner:
    """This is a work-around to the limitation of multiprocessing that the
    function executed in the new module cannot be a method of a class.
    We simply set the instance (self) into the class before forking each
    process, and the class' method calls the instance method for us.
    """
    instance = None

    @classmethod
    def run(cls, index):
        cls.instance._run(index)


def alphadump(d, indent=2, depth=0):
    """Dump a dict to a str,
    with keys in alphabetical order.
    """
    sep = '\n' + ' ' * depth * indent
    return ''.join(
        ("{}: {}{}".format(
            k,
            alphadump(d[k], depth=depth+1) if isinstance(d[k], dict)
            else str(d[k]),
            sep)
         for k in sorted(d.keys()))
    )



"""
Simple "copy" builder.

Copies from one collection to another.
With the optional incremental feature, running twice will only copy the new records, i.e.
running twice in succession will cause the second run to do nothing.

To run:

mgbuild
"""
__author__ = 'Dan Gunter <dkgunter@lbl.gov>'
__date__ = '4/22/14'

from matgendb.builders import core
from matgendb.builders import util
from matgendb.query_engine import QueryEngine

_log = util.get_builder_log("copy")

class CopyBuilder(core.Builder):
    """Copy from one MongoDB collection to another.
    """
    def __init__(self, *args, **kwargs):
        self._target_coll = None
        core.Builder.__init__(self, *args, **kwargs)

    def get_items(self, source=None, target=None, crit=None):
        """Copy records from source to target collection.

        :param source: Input collection
        :type source: QueryEngine
        :param target: Output collection
        :type target: QueryEngine
        :param crit: Filter criteria, e.g. "{ 'flag': True }".
        :type crit: dict
        """
        self._target_coll = target.collection
        if not crit:  # reduce any False-y crit value to None
            crit = None
        cur = source.query(criteria=crit)
        _log.info("source.collection={} crit={} source_records={:d}"
                  .format(source.collection, crit, len(cur)))
        return cur

    def process_item(self, item):
        assert self._target_coll
        self._target_coll.insert(item)




import os
import uuid
import unittest

import bson
import pymongo

from matgendb.query_engine import QueryEngine, QueryResults
from matgendb.tests import common

has_mongo = common.has_mongo()


class QueryResultsTest(unittest.TestCase):
    def setUp(self):
        if has_mongo:
            self.conn = pymongo.MongoClient()
            self.db_name = 'test'
            self.db = self.conn[self.db_name]
            self.coll_name = 'tasks_{}'.format(uuid.uuid4())
            self.coll = self.db[self.coll_name]
            with open(os.path.join(common.TEST_FILES_DIR, 'db_test', 'GaLa.task.json')) as f:
                doc = bson.json_util.loads(f.read())
                self.coll.insert(doc)

    def tearDown(self):
        if has_mongo:
            self.db.drop_collection(self.coll_name)

    @unittest.skipUnless(has_mongo, 'requires MongoDB server')
    def test_queryresult(self):
        qe = QueryEngine(
            connection=self.conn,
            database=self.db_name,
            collection=self.coll_name,
        )
        result = qe.query(
            criteria={'task_id': 'mp-1002133'},
            properties=[
                'calcs_reversed.output.ionic_steps.e_0_energy',
                'calcs_reversed.output.ionic_steps.electronic_steps.e_0_energy',
            ],
        )
        self.assertTrue(isinstance(result, QueryResults))
        print(list(qe.query(criteria={'task_id': 'mp-1002133'})))
        self.assertEqual(len(result), 1)
        doc = list(result)[0]
        self.assertIn('calcs_reversed.output.ionic_steps.e_0_energy', doc)
        v = doc['calcs_reversed.output.ionic_steps.e_0_energy']
        self.assertIsInstance(v, list)
        for elt in v:
            self.assertIsInstance(elt, list)
            for n in elt:
                self.assertIsInstance(n, float)
        self.assertIn('calcs_reversed.output.ionic_steps.electronic_steps.e_0_energy', doc)
        v = doc['calcs_reversed.output.ionic_steps.electronic_steps.e_0_energy']
        for elt in v:
            self.assertIsInstance(elt, list)
            for _elt in elt:
                self.assertIsInstance(_elt, list)
                for n in _elt:
                    self.assertIsInstance(n, float)



"""
Create and access groups of databases,
each configured from different settings.
"""
__author__ = 'Dan Gunter <dkgunter@lbl.gov>'
__date__ = '4/29/14'

import glob
import os
import re
from . import dbconfig, query_engine, util

# aliases
_opj = os.path.join
_opx = os.path.splitext


class CreateQueryEngineError(Exception):
    def __init__(self, cls, settings, err):
        msg = "creating query engine, class={cls} settings={s}: {m}"\
            .format(cls=cls.__name__, s=util.csv_dict(settings), m=err)
        Exception.__init__(self, msg)


class ConfigGroup:
    """Convenient storage and access to a group
    of database configurations.

    Will automatically instantiate these configurations,
    as query engines, on-demand.
    """

    SEP = "."  # Separator between collection names

    def __init__(self, qe_class=query_engine.QueryEngine):
        self._d = RegexDict()   # Main object store
        self._class = qe_class  # Class to used for building QEs
        self._pfx = None        # Prefix to namespace all lookups
        self._cached = {}       # cached QE objs

    def add_path(self, path, pattern="*.json"):
        """Add configuration file(s)
        in `path`. The path can be a single file or a directory.
        If path is a directory, then `pattern`
        (Unix glob-style) will be used to get a list of all config
        files in the directory.

        The name given to each file is the database name
        and collection name (if any) combined with a '.'.

        :param path: File or directory name
        :return: self, for chaining
        """
        if os.path.isdir(path):
            configs = glob.glob(_opj(path, pattern))
        else:
            configs = [path]
        for config in configs:
                cfg = dbconfig.DBConfig(config_file=config)
                cs = cfg.settings
                if dbconfig.DB_KEY not in cs:
                    raise ValueError("No database in '{}'".format(config))
                if dbconfig.COLL_KEY in cs:
                    name = "{}.{}".format(cs[dbconfig.DB_KEY],
                                          cs[dbconfig.COLL_KEY])
                else:
                    name = cs[dbconfig.DB_KEY]
                self.add(name, cfg)
        return self

    def add(self, name, cfg, expand=False):
        """Add a configuration object.

        :param name: Name for later retrieval
        :param cfg: Configuration object
        :param expand: Flag for adding sub-configs for each sub-collection.
                       See discussion in method doc.
        :return: self, for chaining
        :raises: CreateQueryEngineError (only if expand=True)
        """
        self._d[name] = cfg
        if expand:
            self.expand(name)
        return self

    def expand(self, name):
        """Expand config for `name` by adding a sub-configuration for every
        dot-separated collection "below" the given one (or all, if none given).

        For example, for a database 'mydb' with collections
            ['spiderman.amazing', 'spiderman.spectacular', 'spiderman2']
        and a configuration
            {'host':'foo', 'database':'mydb', 'collection':'spiderman'}
        then `expand("mydb.spiderman")` would add keys for 'spiderman.amazing'
        and 'spiderman.spectacular', but *not* 'spiderman2'.

        :param name: Name, or glob-style pattern, for DB configurations.
        :type name: basestring
        :return: None
        :raises: KeyError (if no such configuration)
        """
        if self._is_pattern(name):
            expr = re.compile(self._pattern_to_regex(name))
            for cfg_name in self._d.keys():
                if expr.match(cfg_name):
                    self._expand(cfg_name)
        else:
            self._expand(name)

    def _expand(self, name):
        """Perform real work of `expand()` function."""
        cfg = self._d[name]
        if cfg.collection is None:
            base_coll = ''
        else:
            base_coll = cfg.collection + self.SEP
        qe = self._get_qe(name, cfg)
        coll, db = qe.collection, qe.db
        cur_coll = coll.name
        for coll_name in db.collection_names():
            if coll_name == cur_coll or not coll_name.startswith(base_coll):
                continue
            ex_cfg = cfg.copy()
            ex_cfg.collection = coll_name
            group_name = name + self.SEP + coll_name[len(base_coll):]
            self.add(group_name, ex_cfg, expand=False)

    def uncache(self, name):
        """Remove all created query engines that match `name` from
        the cache (this disconnects from MongoDB, which is the point).

        :param name: Name used for :meth:`add`, or pattern
        :return: None
        """
        delme = []
        if self._is_pattern(name):
            expr = re.compile(self._pattern_to_regex(name))
            for key, obj in self._cached.items():
                if expr.match(key):
                    delme.append(key)
        else:
            if name in self._cached:
                delme.append(name)
        for key in delme:
            del self._cached[key]

    def __getitem__(self, name):
        """Dict-style lookup by name for
        query engine objects. If the input is a pattern,
        this will return a dict where keys are the names and values
        are the QueryEngine objects. Otherwise, will return a
        QueryEngine object. Raises a KeyError if there are no
        results (in either case).

        If this is the first time this query engine
        has been asked for, then it will instantiate the query engine.
        Errors here will raise CreateQueryEngineError.

        :param name: Name to look for; if it
                     ends in '*' then use it as a prefix
                     and return all matching items.
        :return: Single or multiple results (as dict)
        :raises: KeyError, CreateQueryEngineError
        """
        orig_name = name
        if self._pfx is not None:
            name = self._pfx + name
        if self._is_pattern(name):
            name = self._pattern_to_regex(name)
            # fill 'qe' with all items
            qe = {}
            for k, v in self._d.re_get(name).items():
                qe[k] = self._get_qe(k, v)
            if not qe:
                raise KeyError("No configuration found, name='{}' full-regex='{}'"
                               .format(orig_name, name))
        else:
            qe = self._get_qe(name, self._d[name])
        return qe

    def keys(self):
        return self._d.keys()

    def set_prefix(self, prefix=None):
        """Set prefix to use as a namespace for item lookup.
        A dot (.) will be automatically added to the given string.

        :param prefix: Prefix, or None to unset
        :return: None
        """
        if prefix is None:
            self._pfx = None
        else:
            self._pfx = prefix + self.SEP

    @staticmethod
    def _is_pattern(s):
        return s and (s[-1] == '*')

    @staticmethod
    def _pattern_to_regex(pat):
        pat = pat.replace(ConfigGroup.SEP, "\\" + ConfigGroup.SEP)
        return pat[:-1] + ".*"


    def _get_qe(self, key, obj):
        """Instantiate a query engine, or retrieve a cached one.
        """
        if key in self._cached:
            return self._cached[key]
        qe = create_query_engine(obj, self._class)
        self._cached[key] = qe
        return qe


class RegexDict(dict):
    """Extend standard dict to include
    a function that finds values based on a
    regular expression for the key. For example:

       d = RegexDict(tweedledee=1, tweedledum=2)

    """
    def re_keys(self, pattern):
        """Find keys matching `pattern`.

        :param pattern: Regular expression
        :return: Matching keys or empty list
        :rtype: list
        """
        if not pattern.endswith("$"):
            pattern += "$"
        expr = re.compile(pattern)
        return list(filter(expr.match, self.keys()))

    def re_get(self, pattern):
        """Return values whose key matches `pattern`

        :param pattern: Regular expression
        :return: Found values, as a dict.
        """
        return {k: self[k] for k in self.re_keys(pattern)}

def create_query_engine(config, clazz):
    """Create and return new query engine object from the
    given `DBConfig` object.

    :param config: Database configuration
    :type config: dbconfig.DBConfig
    :param clazz: Class to use for creating query engine. Should
                  act like query_engine.QueryEngine.
    :type clazz: class
    :return: New query engine
    """
    try:
        qe = clazz(**config.settings)
    except Exception as err:
        raise CreateQueryEngineError(clazz, config.settings, err)
    return qe



"""
Utility functions used across scripts.
"""
__author__ = "Shyue Ping Ong, Dan Gunter"
__copyright__ = "Copyright 2012-2014, The Materials Project"
__version__ = "1.1"
__maintainer__ = "Dan Gunter"
__email__ = "dkgunter@lbl.gov"
__date__ = "2012-12-01"

## Imports

import bson
import datetime
import json
import logging
from pymongo.mongo_client import MongoClient

from matgendb.dbconfig import DBConfig

# Backwards compatibility from refactor to `dbconfig` module
# Copy of functions that were moved
from matgendb.dbconfig import normalize_auth
# Copy of global constants that were moved
DEFAULT_PORT = DBConfig.DEFAULT_PORT
DEFAULT_CONFIG_FILE = DBConfig.DEFAULT_FILE
DEFAULT_SETTINGS = DBConfig.DEFAULT_SETTINGS

## Logging

_log = logging.getLogger("mg.util")

## Classes

class MongoJSONEncoder(json.JSONEncoder):
    def default(self, o):
        if isinstance(o, bson.objectid.ObjectId):
            return str(o)
        if isinstance(o, datetime.datetime):
            return o.isoformat()
        return json.JSONEncoder.default(self, o)


## Functions

def get_settings(config_file):
    cfg = DBConfig(config_file)
    return cfg.settings


def get_database(config_file=None, settings=None, admin=False, **kwargs):
    d = get_settings(config_file) if settings is None else settings
    conn = MongoClient(host=d["host"], port=d["port"], **kwargs)
    db = conn[d["database"]]
    try:
        user = d["admin_user"] if admin else d["readonly_user"]
        passwd = d["admin_password"] if admin else d["readonly_password"]
        db.authenticate(user, passwd)
    except (KeyError, TypeError, ValueError):
        _log.warn("No {admin,readonly}_user/password found in config. file, "
                  "accessing DB without authentication")
    return db


def get_collection(config_file, admin=False, settings=None):
    if settings is None:
        settings = get_settings(config_file)
    db = get_database(admin=admin, settings=settings)
    return db[settings["collection"]]


def collection_keys(coll, sep='.'):
    """Get a list of all (including nested) keys in a collection.
    Examines the first document in the collection.

    :param sep: Separator for nested keys
    :return: List of str
    """
    def _keys(x, pre=''):
        for k in x:
            yield (pre + k)
            if isinstance(x[k], dict):
                for nested in _keys(x[k], pre + k + sep):
                    yield nested

    return list(_keys(coll.find_one()))

def csv_list(l):
    """Format list to a string with comma-separated values.
    """
    if len(l) == 0:
        return ""
    return ', '.join(map(str, l))


def quotable(v):
    if isinstance(v, int) or isinstance(v, float):
        return str(v)
    return "'{}'".format(v)


def csv_dict(d):
    """Format dict to a string with comma-separated values.
    """
    if len(d) == 0:
        return "{}"
    return "{" + ', '.join(["'{}': {}".format(k, quotable(v))
                            for k, v in d.items()]) + "}"

def kvp_dict(d):
    """Format dict to key=value pairs.
    """
    return ', '.join(
        ["{}={}".format(k, quotable(v)) for k, v in d.items()])




"""
A helper script for many matgendb functions.
"""

__author__ = "Shyue Ping Ong"
__copyright__ = "Copyright 2012, The Materials Project"
__version__ = "1.2"
__maintainer__ = "Shyue Ping Ong"
__email__ = "shyue@mit.edu"
__date__ = "Dec 1, 2012"

import datetime
import logging
import multiprocessing
import json
import sys
import argparse
import six

from pymongo import MongoClient, ASCENDING

from pymatgen.apps.borg.queen import BorgQueen

from matgendb import SETTINGS
from matgendb.query_engine import QueryEngine
from matgendb.creator import VaspToDbTaskDrone
from matgendb.dbconfig import DBConfig
from matgendb.util import get_settings, DEFAULT_SETTINGS, MongoJSONEncoder

_log = logging.getLogger("mg")  # parent


def init_db(args):
    settings = DBConfig.ALL_SETTINGS
    defaults = dict(DEFAULT_SETTINGS)
    doc = {}
    print("Please supply the following configuration values")
    print("(press Enter if you want to accept the defaults)\n")
    for k in settings:
        v = defaults.get(k, '')
        val = six.moves.input("Enter {} (default: {}) : ".format(k, v))
        doc[k] = val if val else v
    doc["port"] = int(doc["port"])  # enforce the port as an int
    for k in ["admin_user", "admin_password", "readonly_user",
              "readonly_password"]:
        v = defaults.get(k, '')
        val = six.moves.input("Enter {} (default: {}) : ".format(k, v))
        doc[k] = val if val else v
    with open(args.config_file, "w") as f:
        json.dump(doc, f, indent=4, sort_keys=True)
    print("\nConfiguration written to {}!".format(args.config_file))


def update_db(args):
    FORMAT = "%(relativeCreated)d msecs : %(message)s"

    if args.logfile:
        logging.basicConfig(level=logging.INFO, format=FORMAT,
                            filename=args.logfile[0])
    else:
        logging.basicConfig(level=logging.INFO, format=FORMAT)

    d = get_settings(args.config_file)

    _log.info("Db insertion started at {}.".format(datetime.datetime.now()))
    additional_fields = {"author": args.author, "tags": args.tag}
    drone = VaspToDbTaskDrone(
        host=d["host"], port=d["port"],  database=d["database"],
        user=d["admin_user"], password=d["admin_password"],
        parse_dos=args.parse_dos,
        collection=d["collection"], update_duplicates=args.force_update_dupes,
        additional_fields=additional_fields, mapi_key=d.get("mapi_key", None))
    ncpus = multiprocessing.cpu_count() if not args.ncpus else args.ncpus
    _log.info("Using {} cpus...".format(ncpus))
    queen = BorgQueen(drone, number_of_drones=ncpus)
    queen.parallel_assimilate(args.directory)
    tids = list(map(int, filter(lambda x: x, queen.get_data())))
    _log.info("Db upate completed at {}.".format(datetime.datetime.now()))
    _log.info("{} new task ids inserted.".format(len(tids)))


def optimize_indexes(args):
    d = get_settings(args.config_file)
    c = MongoClient(d["host"], d["port"])
    db = c[d["database"]]
    db.authenticate(d["admin_user"], d["admin_password"])
    coll = db[d["collection"]]
    coll.drop_indexes()
    coll.ensure_index('task_id', unique=True)
    for key in ['unit_cell_formula', 'reduced_cell_formula', 'chemsys',
                'nsites', 'pretty_formula', 'analysis.e_above_hull',
                "icsd_ids"]:
        print("Building {} index".format(key))
        coll.ensure_index(key)
    print("Building nelements and elements compound index")
    compound_index = [('nelements', ASCENDING), ('elements', ASCENDING)]
    coll.ensure_index(compound_index)
    coll.ensure_index(compound_index)


def query_db(args):
    from tabulate import tabulate
    d = get_settings(args.config_file)
    qe = QueryEngine(host=d["host"], port=d["port"], database=d["database"],
                     user=d["readonly_user"], password=d["readonly_password"],
                     collection=d["collection"],
                     aliases_config=d.get("aliases_config", None))
    criteria = None
    if args.criteria:
        try:
            criteria = json.loads(args.criteria)
        except ValueError:
            print("Criteria {} is not a valid JSON string!".format(
                args.criteria))
            sys.exit(-1)

    # TODO: document this 'feature' --dang 4/4/2013
    is_a_file = lambda s: len(s) == 1 and s[0].startswith(':')
    if is_a_file(args.properties):
        with open(args.properties[0][1:], 'rb') as f:
            props = [s.strip() for s in f]
    else:
        props = args.properties

    if args.dump_json:
        for r in qe.query(properties=props, criteria=criteria):
            print(json.dumps(r, cls=MongoJSONEncoder))
    else:
        t = []
        for r in qe.query(properties=props, criteria=criteria):
            t.append([r[p] for p in props])
        print(tabulate(t, headers=props))


class StatsCommands(object):
    """Encapsulate statistics commands.
    """
    #: Constants for names of format options
    FORMAT_YAML, FORMAT_CLEAN = "yaml", "simple"

    def __init__(self, query_engine, output_format=None, latest_prop=None, **extra_kw):
        """Constructor.

        :param query_engine: Connection to Mongo
        :type query_engine: matgendb.query_engine.QueryEngine
        """
        self._qe = query_engine
        self._db, self._coll = query_engine.db, query_engine.collection
        self.indent = ' ' * 4
        # Keywords
        self._latest = latest_prop
        if output_format == self.FORMAT_YAML:
            self._format = self.yaml_format
            self._show_header = True
        elif output_format == self.FORMAT_CLEAN:
            self._format = self.clean_format
            self._show_header = False

    def header(self):
        if not self._show_header:
            return ""
        s = self._format((("Database", self._db.name),
                          ("Collection", self._coll.name),
                          ("Values", "")))
        return s

    def stat(self, name):
        func = 'stat_{}'.format(name)
        return getattr(self, func)()

    def stat_latest(self):
        cur = self._coll.find({}, [self._latest])\
                        .sort(self._latest, -1)\
                        .limit(1)
        try:
            value = "{}".format(cur[0][self._latest])
        except IndexError:
            _log.error("Cannot get latest value: no records with field '{}'"
                       .format(self._latest))
            return ""
        return self._format([("Latest", value)], depth=1)

    def stat_count(self):
        value = "{:d}".format(self._coll.count())
        return self._format([("Count", value)], depth=1)

    def yaml_format(self, data, depth=0):
        rows, ind = [], self.indent * depth
        for k, v in data:
            rows.append("{i}{k}: {v}".format(i=ind, k=k, v=v))
        return '\n'.join(rows) + "\n"

    def clean_format(self, data, depth=0):
        rows = ["{} {}".format(d[0].lower(), d[1]) for d in data]
        return '\n'.join(rows) + "\n"


def db_stats(args):
    """Database and/or collection statistics.
    """
    # Init db.
    cfg = get_settings(args.config_file)
    normalize_userpass(cfg)
    qe = QueryEngine(**cfg)

    # Copy stat args and keywords.
    stats, kw = {}, {}
    for k, v in vars(args).iteritems():
        if k.startswith("st_"):
            stats[k[3:]] = v
        else:
            kw[k] = v

    # Process 'show all'.
    if stats['showall']:
        for k in stats.iterkeys():
            stats[k] = True
    del stats['showall']

    # Output destination.
    w = sys.stdout

    # Run commands.
    cmd = StatsCommands(qe, **kw)
    w.write(cmd.header())
    for k in stats.iterkeys():
        if stats[k]:
            w.write(cmd.stat(k))


def normalize_userpass(cfg):
    """In DB conn. config, normalize user/password from readonly and admin prefixes.
    In the end, there will be only keys 'user' and 'password'.
    """
    for pfx in 'readonly', 'admin':  # in reverse order of priority, to overwrite
        if (pfx + '_user') in cfg and (pfx + '_password') in cfg:
            cfg[QueryEngine.USER_KEY] = cfg[pfx + '_user']
            cfg[QueryEngine.PASSWORD_KEY] = cfg[pfx + '_password']
            del cfg[pfx + '_user']
            del cfg[pfx + '_password']


def init_logging(args):
    """Initialize verbosity
    """
    _log.propagate = False
    hndlr = logging.StreamHandler()
    hndlr.setFormatter(logging.Formatter("[%(levelname)-6s] %(asctime)s %(name)s :: %(message)s"))
    _log.addHandler(hndlr)
    if 'quiet' in args and args.quiet:
        lvl = logging.CRITICAL
    else:
        vb = args.vb if 'vb' in args else 0
        # Level:  default      -v            -vv
        lvl = (logging.WARN, logging.INFO, logging.DEBUG)[min(vb, 2)]
    _log.setLevel(lvl)


if __name__ == "__main__":
    db_file = SETTINGS.get("PMGDB_DB_FILE")
    parser = argparse.ArgumentParser(description="""
    mgdb is a complete command line db management script for pymatgen-db. It
    provides the facility to insert vasp runs, perform queries, and run a web
    server for exploring databases that you create. Type mgdb -h to see the
    various options.

    Author: Shyue Ping Ong
    Version: 3.0
    Last updated: Mar 23 2013""")

    # Parents for all subparsers.
    parent_vb = argparse.ArgumentParser(add_help=False)
    parent_vb.add_argument('--quiet', '-q', dest='quiet', action="store_true",
                           default=False,
                           help="Minimal verbosity.")
    parent_vb.add_argument('--verbose', '-v', dest='vb', action="count",
                           default=0,
                           help="Print more verbose messages to standard error. "
                                "Repeatable. (default=ERROR)")
    parent_cfg = argparse.ArgumentParser(add_help=False)
    parent_cfg.add_argument("-c", "--config", dest="config_file", type=str,
                            default=db_file,
                            help="Config file to use. Generate one using mgdb "
                            "init --config filename.json if necessary. "
                            "Otherwise, the code searches for a db.json. If"
                            "none is found, an no-authentication "
                            "localhost:27017/vasp database and tasks "
                            "collection is assumed.")

    # change db_file to the default "db.json" if it does not exist
    db_file = db_file or "db.json"

    # Init for all subparsers.
    subparsers = parser.add_subparsers()

    # The 'init' subcommand.
    pinit = subparsers.add_parser("init", help="Initialization tools.",
                                  parents=[parent_vb])
    pinit.add_argument("-c", "--config", dest="config_file", type=str,
                       nargs='?', default=db_file,
                       help="Creates an db config file for the database. "
                            "Default filename is db.json.")
    pinit.set_defaults(func=init_db)

    popt = subparsers.add_parser("optimize", help="Optimization tools.")

    popt.add_argument("-c", "--config", dest="config_file", type=str,
                      nargs='?', default=db_file,
                      help="Creates an db config file for the database. "
                           "Default filename is db.json.")
    popt.set_defaults(func=optimize_indexes)

    # The 'insert' subcommand.
    pinsert = subparsers.add_parser("insert", help="Insert vasp runs.",
                                    parents=[parent_vb, parent_cfg])
    pinsert.add_argument("directory", metavar="directory", type=str,
                         default=".", help="Root directory for runs.")
    pinsert.add_argument("-l", "--logfile", dest="logfile", type=str,
                         help="File to log db insertion. Defaults to stdout.")
    pinsert.add_argument("-t", "--tag", dest="tag", type=str, nargs="+",
                         default=[],
                         help="Tag your runs for easier search."
                              " Accepts multiple space-separated tags. E.g.,"
                              " '--tag metal energy'")
    pinsert.add_argument("-f", "--force", dest="force_update_dupes",
                         action="store_true",
                         help="Force update duplicates. This forces the "
                              "analyzer to reanalyze already inserted data.")
    pinsert.add_argument("-d", "--parse_dos", dest="parse_dos",
                         action="store_true",
                         help="Whether to parse the dos.")
    pinsert.add_argument("-a", "--author", dest="author", type=str, nargs=1,
                         default=None,
                         help="Enter a *unique* author field so that you can "
                              "trace back what you ran.")
    pinsert.add_argument("-n", "--ncpus", dest="ncpus", type=int,
                         default=None,
                         help="Number of CPUs to use in inserting. If "
                              "not specified, multiprocessing will use "
                              "the number of cpus detected.")
    pinsert.set_defaults(func=update_db)

    # The 'query' subcommand.
    pquery = subparsers.add_parser("query",
                                   help="Query tools. Requires the "
                                        "use of pretty_table.",
                                   parents=[parent_vb, parent_cfg])
    pquery.add_argument("--crit", dest="criteria", type=str, default=None,
                        help="Query criteria in typical json format. E.g., "
                             "{\"task_id\": 1}.")
    pquery.add_argument("--props", dest="properties", type=str, default=[],
                        nargs='+', required=True,
                        help="Desired properties. Repeatable. E.g., pretty_formula, "
                             "task_id, energy...")
    pquery.add_argument("--dump", dest="dump_json", action='store_true', default=False,
                        help="Simply dump results to JSON instead of a tabular view")
    pquery.set_defaults(func=query_db)

    # The 'stats' subcommand.
    pstats = subparsers.add_parser("stats", help="Database and/or collection statistics.",
                                   parents=[parent_vb, parent_cfg])
    pstats.add_argument("--count", help="Show count of records", dest="st_count", action="store_true")
    pstats.add_argument("--latest", help="Show time of latest record", dest="st_latest", action="store_true")
    #pstats.add_argument("--size", help="Show min/max/avg/total record sizes", dest="", action="store_true")
    pstats.add_argument("--all", help="Show all stats", dest="st_showall", action="store_true")
    pstats.add_argument("--latest-prop", help="Property to use for latest record (default='updated_at')",
                        default="updated_at", dest="latest_prop")
    pstats.add_argument("-f", "--format", help="Output format", choices=[StatsCommands.FORMAT_YAML,
                                                                         StatsCommands.FORMAT_CLEAN],
                        default="yaml", dest="output_format")
    pstats.set_defaults(func=db_stats)

    # Parse args
    args = parser.parse_args()

    init_logging(args)

    # Run appropriate subparser function.
    args.func(args)



"""
Database configuration functions.

Main class is DBConfig, which encapsulates a database configuration
passed in as a file or object. For example::

    cfg1 = DBConfig()  # use defaults
    cfg2 = DBConfig("/path/to/myfile.json")  # read from file
    f = open("/other/file.json")
    cfg3 = DBConfig(f)  # read from file object
    # access dict of parsed conf. settings
    settings = cfg1.settings

"""
__author__ = 'Dan Gunter <dkgunter@lbl.gov>'
__date__ = '4/25/14'

import os
import yaml

# Constants for keys
HOST_KEY = "host"
PORT_KEY = "port"
DB_KEY = "database"
COLL_KEY = "collection"
USER_KEY = "user"
PASS_KEY = "password"
ALIASES_KEY = "aliases"

class ConfigurationFileError(Exception):
    def __init__(self, filename, err):
        msg = "reading '{}': {}".format(filename, err)
        Exception.__init__(self, msg)

class DBConfig:
    """Database configuration.
    """

    DEFAULT_PORT = 27017
    DEFAULT_FILE = 'db.json'
    ALL_SETTINGS = [
        HOST_KEY,
        PORT_KEY,
        DB_KEY,
        COLL_KEY,
        ALIASES_KEY,
    ]
    DEFAULT_SETTINGS = [
        (HOST_KEY, "localhost"),
        (PORT_KEY, DEFAULT_PORT),
        (DB_KEY, "vasp"),
        (ALIASES_KEY, {}),
    ]

    def __init__(self, config_file=None, config_dict=None):
        """
        Constructor.

        Settings are created from config_dict, if given,
        or parsed config_file, if given, otherwise
        the DEFAULT_FILE is tried and if that is not present
        the DEFAULT_SETTINGS are used without modification.

        :param config_file: Read configuration from this file.
        :type config_file: file or str path
        :param config_dict: Set configuration from this dictionary.
        :raises: ConfigurationFileError if cannot read/parse config_file
        """
        self._cfg = dict(self.DEFAULT_SETTINGS)
        settings = {}
        if config_dict:
            settings = config_dict.copy()
            auth_aliases(settings)
        else:
            # Try to use DEFAULT_FILE if no config_file
            if config_file is None:
                if os.path.exists(self.DEFAULT_FILE):
                    config_file = self.DEFAULT_FILE
            # If there was a config_file, parse it
            if config_file is not None:
                try:
                    settings = get_settings(config_file)
                except Exception as err:
                    path = _as_file(config_file).name
                    raise ConfigurationFileError(path, err)
        self._cfg.update(settings)
        normalize_auth(self._cfg)

    def __str__(self):
        return str(self._cfg)

    def copy(self):
        """Return a copy of self (internal settings are copied).
        """
        return DBConfig(config_dict=self._cfg.copy())

    @property
    def settings(self):
        return self._cfg

    @property
    def host(self):
        return self._cfg.get(HOST_KEY, None)

    @property
    def port(self):
        return self._cfg.get(PORT_KEY, self.DEFAULT_PORT)

    @property
    def dbname(self):
        """Name of the database."""
        return self._cfg.get(DB_KEY, None)

    @dbname.setter
    def dbname(self, value):
        self._cfg[DB_KEY] = value

    @property
    def collection(self):
        return self._cfg.get(COLL_KEY, None)

    @collection.setter
    def collection(self, value):
        self._cfg[COLL_KEY] = value

    @property
    def user(self):
        return self._cfg.get(USER_KEY, None)

    @property
    def password(self):
        return self._cfg.get(PASS_KEY, None)


def get_settings(infile):
    """Read settings from input file.

    :param infile: Input file for JSON settings.
    :type infile: file or str path
    :return: Settings parsed from file
    :rtype: dict
    """
    settings = yaml.load(_as_file(infile))
    if not hasattr(settings, 'keys'):
        raise ValueError("Settings not found in {}".format(infile))

    # Processing of namespaced parameters in .pmgrc.yaml.
    processed_settings = {}
    for k, v in settings.items():
        if k.startswith("PMG_DB_"):
            processed_settings[k[7:].lower()] = v
        else:
            processed_settings[k] = v
    auth_aliases(processed_settings)
    return processed_settings

def auth_aliases(d):
    """Interpret user/password aliases.
    """
    for alias, real in ((USER_KEY, "readonly_user"),
                        (PASS_KEY, "readonly_password")):
        if alias in d:
            d[real] = d[alias]
            del d[alias]

def normalize_auth(settings, admin=True, readonly=True, readonly_first=False):
    """Transform the readonly/admin user and password to simple user/password,
    as expected by QueryEngine. If return value is true, then
    admin or readonly password will be in keys "user" and "password".

    :param settings: Connection settings
    :type settings: dict
    :param admin: Check for admin password
    :param readonly: Check for readonly password
    :param readonly_first: Check for readonly password before admin
    :return: Whether user/password were found
    :rtype: bool
    """
    U, P = USER_KEY, PASS_KEY
    # If user/password, un-prefixed, exists, do nothing.
    if U in settings and P in settings:
        return True

    # Set prefixes
    prefixes = []
    if readonly_first:
        if readonly:
            prefixes.append("readonly_")
        if admin:
            prefixes.append("admin_")
    else:
        if admin:
            prefixes.append("admin_")
        if readonly:
            prefixes.append("readonly_")

    # Look for first user/password matching.
    found = False
    for pfx in prefixes:
        ukey, pkey = pfx + U, pfx + P
        if ukey in settings and pkey in settings:
            settings[U] = settings[ukey]
            settings[P] = settings[pkey]
            found = True
            break

    return found


def _as_file(f, mode='r'):
    if isinstance(f, str):
        return open(f, mode)
    return f



"""
Common utility functions for the database functions
"""
__author__ = 'Dan Gunter <dkgunter@lbl.gov>'
__date__ = '11/4/13'

## Imports
import logging
import matgendb
# Stdlib
import os
# Local
from matgendb.util import csv_dict, csv_list

## Globals

## Functions and Classes


_top_dir = os.path.dirname(os.path.abspath(matgendb.__file__))


def get_test_dir(name=None):
    """Get path to subdirectory with test files related
    to a given class.

    :param name: Name of class
    :return: Path
    :rtype: str
    """

    subdir = name + '_test'
    return os.path.join(_top_dir, '..', 'test_files', subdir)


def get_schema_dir(db_version=1):
    """Get path to directory with schemata.

    :param db_version: Version of the database
    :type db_version: int
    :return: Path
    :rtype: str
    """
    v = str(db_version)
    return os.path.join(_top_dir, '..', 'schemata', 'versions', v)


def get_schema_file(db_version=1, db="mg_core", collection="materials"):
    """Get file with appropriate schema.

    :param db_version: Version of the database
    :type db_version: int
    :param db: Name of database, e.g. 'mg_core'
    :type db: str
    :param collection: Name of collection, e.g. 'materials'
    :type collection: str
    :return: File with schema
    :rtype: file
    :raise: IOError, if file is not found or not accessible
    """
    d = get_schema_dir(db_version=db_version)
    schemafile = "{}.{}.json".format(db, collection)
    f = open(os.path.join(d, schemafile), "r")
    return f

def get_builder_log(name):
    """Get a logging object, in the right place in the
    hierarchy, for a given builder.

    :param name: Builder name, e.g. 'my_builder'
    :type name: str
    :returns: New logger
    :rtype: logging.Logger
    """
    return logging.getLogger("mg.builders." + name)





#!/usr/bin/env python

"""
This module implements a version of pymatgen's Transmuter to generate
TransformedStructures from DB data sources. They enable the
high-throughput generation of new structures and input files.
"""

from __future__ import division


__author__ = "Shyue Ping Ong"
__copyright__ = "Copyright 2012, The Materials Project"
__version__ = "0.1"
__maintainer__ = "Shyue Ping Ong"
__email__ = "shyue@mit.edu"
__date__ = "Mar 4, 2012"

import datetime
from pymatgen.alchemy.materials import TransformedStructure
from pymatgen.alchemy.transmuters import StandardTransmuter


class QeTransmuter(StandardTransmuter):
    """
    The QeTransmuter uses a QueryEngine to retrieve and generate new structures
    from a database.
    """

    def __init__(self, queryengine, criteria, transformations,
                 extend_collection=0, ncores=None):
        """Constructor.

        Args:
            queryengine:
                QueryEngine object for database access
            criteria:
                A criteria to search on, which is passed to queryengine's
                get_entries method.
            transformations:
                New transformations to be applied to all structures
            extend_collection:
                Whether to use more than one output structure from one-to-many
                transformations. extend_collection can be a number, which
                determines the maximum branching for each transformation.
            ncores:
                Number of cores to use for applying transformations.
                Uses multiprocessing.Pool
        """
        entries = queryengine.get_entries(criteria, inc_structure=True)

        source = "{}:{}/{}/{}".format(queryengine.host, queryengine.port,
                                      queryengine.database_name,
                                      queryengine.collection_name)

        def get_history(entry):
            return [{"source": source,
                     "criteria": criteria,
                     "entry": entry.as_dict(),
                     "datetime": datetime.datetime.utcnow()}]

        transformed_structures = [TransformedStructure(
            entry.structure, [], history=get_history(entry))
            for entry in entries]
        StandardTransmuter.__init__(self, transformed_structures,
                                    transformations=transformations,
                                    extend_collection=extend_collection,
                                    ncores=ncores)



"""
This module provides a QueryEngine that simplifies queries for Mongo databases
generated using hive.
"""

from __future__ import division

__author__ = "Shyue Ping Ong, Michael Kocher, Dan Gunter"
__copyright__ = "Copyright 2011, The Materials Project"
__version__ = "2.0"
__maintainer__ = "Shyue Ping Ong"
__email__ = "shyuep@gmail.com"
__status__ = "Production"
__date__ = "Mar 2 2013"

import json
import itertools
import logging
import os
import gridfs
import zlib
from collections import OrderedDict, Iterable

import pymongo
from pymongo import MongoClient
from pymatgen import Structure, Composition
from pymatgen.electronic_structure.core import Orbital, Spin
from pymatgen.electronic_structure.dos import CompleteDos, Dos
from pymatgen.entries.computed_entries import ComputedEntry,\
    ComputedStructureEntry

_log = logging.getLogger('mg.' + __name__)


class QueryEngine:
    """This class defines a QueryEngine interface to a Mongo Collection based on
    a set of aliases. This query engine also provides convenient translation
    between various pymatgen objects and database objects.

    The major difference between the QueryEngine's query() method and pymongo's
    find() method is the treatment of nested fields. QueryEngine's query
    will map the final result to a root level string, while pymmongo will
    return the doc as is. For example, let's say you have a document
    that is of the following form::

        {"a": {"b" : 1}}

    Using pymongo.find({}, fields=["a.b"]), you will get a doc where you need
    to do doc["a"]["b"] to access the final result (1). Using
    QueryEngine.query(properties=["a.b"], you will obtain a result that can be
    accessed simply as doc["a.b"].
    """

    # avoid hard-coding these in other places
    ALIASES_CONFIG_KEY = 'aliases_config'
    COLLECTION_KEY = 'collection'
    HOST_KEY = 'host'
    PORT_KEY = 'port'
    DB_KEY = 'database'
    USER_KEY = 'user'
    PASSWORD_KEY = 'password'

    # Aliases and defaults
    aliases = None            #: See `aliases` arg to constructor
    default_criteria = None   #: See `default_criteria` arg to constructor
    default_properties = None #: See `default_properties` arg to constructor
    # Post-processing operations
    query_post = None         #: See `query_post` arg to constructor
    result_post = None        #: See `result_post` arg to constructor

    def __init__(self, host="127.0.0.1", port=27017, database="vasp",
                 user=None, password=None, collection="tasks",
                 aliases_config=None, default_properties=None,
                 query_post=None, result_post=None,
                 connection=None, replicaset=None, **ignore):
        """Constructor.

        Args:
            host (str): Hostname of database machine.
            port (int): Port for db access.
            database (str): Name of database to access.
            user (str): User for db access. `None` means no authentication.
            password (str): Password for db access. `None` means no auth.
            collection (str): Collection to query. Defaults to "tasks".
            connection (pymongo.Connection): If given, ignore 'host' and 'port'
                and use existing connection.
            aliases_config(dict):
                An alias dict to use. Defaults to None, which means the default
                aliases defined in "aliases.json" is used. The aliases config
                should be of the following format::

                    {
                        "aliases": {
                            "e_above_hull": "analysis.e_above_hull",
                            "energy": "output.final_energy",
                            ....
                        },
                        "defaults": {
                            "state": "successful"
                        }
                    }
                aliases (dict): Keys are the incoming property, values are the
                    property it will be translated to. This makes it easier
                    to organize the doc format in a way that is different from the
                    query format.
                defaults (dict): Criteria that should be applied
                    by default to all queries. For example, a collection may
                    contain data from both successful and unsuccessful runs but
                    for most querying purposes, you may want just successful runs
                    only. Note that defaults do not affect explicitly specified
                    criteria, i.e., if you suppy a query for {"state": "killed"},
                    this will override the default for {"state": "successful"}.
            default_properties (list): Property names (strings) to use by
                default, if no `properties` are given to query().
            query_post (list): Functions to post-process the `criteria` passed
                to `query()`, after aliases are resolved.
                Function takes two args, the criteria dict and list of
                result properties. Both may be modified in-place.
            result_post (list): Functions to post-process the cursor records.
                Function takes one arg, the document for the current record,
                that is modified in-place.
        """
        self.host = host
        self.port = port
        self.replicaset = replicaset
        self.database_name = database
        if connection is None:
            # can't pass replicaset=None to MongoClient (fails validation)
            if self.replicaset:
                self.connection = MongoClient(self.host, self.port,
                                              replicaset=self.replicaset)
            else:
                self.connection = MongoClient(self.host, self.port)
        else:
            self.connection = connection
        self.db = self.connection[database]
        if user:
            self.db.authenticate(user, password)
        self.collection_name = collection
        self.set_aliases_and_defaults(aliases_config=aliases_config,
                                      default_properties=default_properties)
        # Post-processing functions
        self.query_post = query_post or []
        self.result_post = result_post or []

    @property
    def collection_name(self):
        return self._collection_name

    @collection_name.setter
    def collection_name(self, value):
        """Switch to another collection.
        Note that you may have to set the aliases and default properties if the
        schema of the new collection differs from the current collection.
        """
        self._collection_name = value
        self.collection = self.db[value]

    def set_collection(self, collection):
        """
        To Be Deprecated: this should be replaced with the @property collection_name

        Switch to another collection. Note that you may have to set the
        aliases and default properties via set_aliases_and_defaults if the
        schema of the new collection differs from the current collection.

        Args:
            collection:
                Name of collection.
        """
        # Use the @property setter collection_name setter above
        self.collection_name = collection


    def set_aliases_and_defaults(self, aliases_config=None,
                                 default_properties=None):
        """
        Set the alias config and defaults to use. Typically used when
        switching to a collection with a different schema.

        Args:
            aliases_config:
                An alias dict to use. Defaults to None, which means the default
                aliases defined in "aliases.json" is used. See constructor
                for format.
            default_properties:
                List of property names (strings) to use by default, if no
                properties are given to the 'properties' argument of
                query().
        """
        if aliases_config is None:
            with open(os.path.join(os.path.dirname(__file__),
                                   "aliases.json")) as f:
                d = json.load(f)
                self.aliases = d.get("aliases", {})
                self.default_criteria = d.get("defaults", {})
        else:
            self.aliases = aliases_config.get("aliases", {})
            self.default_criteria = aliases_config.get("defaults", {})
        # set default properties
        if default_properties is None:
            self._default_props, self._default_prop_dict = None, None
        else:
            self._default_props, self._default_prop_dict = \
                self._parse_properties(default_properties)


    def __enter__(self):
        """Allows for use with the 'with' context manager"""
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """Allows for use with the 'with' context manager"""
        self.close()

    def close(self):
        """Disconnects the connection."""
        self.connection.disconnect()

    def get_entries_in_system(self, elements, inc_structure=False,
                              optional_data=None, additional_criteria=None):
        """
        Gets all entries in a chemical system, e.g. Li-Fe-O will return all
        Li-O, Fe-O, Li-Fe, Li-Fe-O compounds.

        .. note::

            The get_entries_in_system and get_entries  methods should be used
            with care. In essence, all entries, GGA, GGA+U or otherwise,
            are returned.  The dataset is very heterogeneous and not
            directly comparable.  It is highly recommended that you perform
            post-processing using pymatgen.entries.compatibility.

        Args:
            elements:
                Sequence of element symbols, e.g. ['Li','Fe','O']
            inc_structure:
                Optional parameter as to whether to include a structure with
                the ComputedEntry. Defaults to False. Use with care - including
                structures with a large number of entries can potentially slow
                down your code to a crawl.
            optional_data:
                Optional data to include with the entry. This allows the data
                to be access via entry.data[key].
            additional_criteria:
                Added ability to provide additional criteria other than just
                the chemical system.

        Returns:
            List of ComputedEntries in the chemical system.
        """
        chemsys_list = []
        for i in range(len(elements)):
            for combi in itertools.combinations(elements, i + 1):
                chemsys = "-".join(sorted(combi))
                chemsys_list.append(chemsys)
        crit = {"chemsys": {"$in": chemsys_list}}
        if additional_criteria is not None:
            crit.update(additional_criteria)
        return self.get_entries(crit, inc_structure,
                                optional_data=optional_data)

    def get_entries(self, criteria, inc_structure=False, optional_data=None):
        """
        Get ComputedEntries satisfying a particular criteria.

        .. note::

            The get_entries_in_system and get_entries  methods should be used
            with care. In essence, all entries, GGA, GGA+U or otherwise,
            are returned.  The dataset is very heterogeneous and not
            directly comparable.  It is highly recommended that you perform
            post-processing using pymatgen.entries.compatibility.

        Args:
            criteria:
                Criteria obeying the same syntax as query.
            inc_structure:
                Optional parameter as to whether to include a structure with
                the ComputedEntry. Defaults to False. Use with care - including
                structures with a large number of entries can potentially slow
                down your code to a crawl.
            optional_data:
                Optional data to include with the entry. This allows the data
                to be access via entry.data[key].

        Returns:
            List of pymatgen.entries.ComputedEntries satisfying criteria.
        """
        all_entries = list()
        optional_data = [] if not optional_data else list(optional_data)
        optional_data.append("oxide_type")
        fields = [k for k in optional_data]
        fields.extend(["task_id", "unit_cell_formula", "energy", "is_hubbard",
                       "hubbards", "pseudo_potential.labels",
                       "pseudo_potential.functional", "run_type",
                       "input.is_lasph", "input.xc_override",
                       "input.potcar_spec"])
        if inc_structure:
            fields.append("output.crystal")

        for c in self.query(fields, criteria):
            func = c["pseudo_potential.functional"]
            labels = c["pseudo_potential.labels"]
            symbols = ["{} {}".format(func, label) for label in labels]
            parameters = {"run_type": c["run_type"],
                          "is_hubbard": c["is_hubbard"],
                          "hubbards": c["hubbards"],
                          "potcar_symbols": symbols,
                          "is_lasph": c.get("input.is_lasph") or False,
                          "potcar_spec": c.get("input.potcar_spec"),
                          "xc_override": c.get("input.xc_override")}
            optional_data = {k: c[k] for k in optional_data}
            if inc_structure:
                struct = Structure.from_dict(c["output.crystal"])
                entry = ComputedStructureEntry(struct, c["energy"],
                                               0.0, parameters=parameters,
                                               data=optional_data,
                                               entry_id=c["task_id"])
            else:
                entry = ComputedEntry(Composition(c["unit_cell_formula"]),
                                      c["energy"], 0.0, parameters=parameters,
                                      data=optional_data,
                                      entry_id=c["task_id"])
            all_entries.append(entry)

        return all_entries

    def _parse_criteria(self, criteria):
        """
        Internal method to perform mapping of criteria to proper mongo queries
        using aliases, as well as some useful sanitization. For example, string
        formulas such as "Fe2O3" are auto-converted to proper mongo queries of
        {"Fe":2, "O":3}.

        If 'criteria' is None, returns an empty dict. Putting this logic here
        simplifies callers and allows subclasses to insert something even
        when there are no criteria.
        """
        if criteria is None:
            return dict()
        parsed_crit = dict()
        for k, v in self.default_criteria.items():
            if k not in criteria:
                parsed_crit[self.aliases.get(k, k)] = v

        for key, crit in list(criteria.items()):
            if key in ["normalized_formula", "reduced_cell_formula"]:
                comp = Composition(crit)
                parsed_crit["pretty_formula"] = comp.reduced_formula
            elif key == "unit_cell_formula":
                comp = Composition(crit)
                crit = comp.as_dict()
                for el, amt in crit.items():
                    parsed_crit["{}.{}".format(self.aliases[key], el)] = amt
                parsed_crit["nelements"] = len(crit)
                parsed_crit['pretty_formula'] = comp.reduced_formula
            elif key in ["$or", "$and"]:
                parsed_crit[key] = [self._parse_criteria(m) for m in crit]
            else:
                parsed_crit[self.aliases.get(key, key)] = crit
        return parsed_crit

    def ensure_index(self, key, unique=False):
        """Wrapper for pymongo.Collection.ensure_index
        """
        return self.collection.ensure_index(key, unique=unique)

    def query(self, properties=None, criteria=None, distinct_key=None,
              **kwargs):
        """
        Convenience method for database access.  All properties and criteria
        can be specified using simplified names defined in Aliases.  You can
        use the supported_properties property to get the list of supported
        properties.

        Results are returned as an iterator of dicts to ensure memory and cpu
        efficiency.

        Note that the dict returned have keys also in the simplified names
        form, not in the mongo format. For example, if you query for
        "analysis.e_above_hull", the returned result must be accessed as
        r['analysis.e_above_hull'] instead of mongo's
        r['analysis']['e_above_hull']. This is a *feature* of the query engine
        to allow simple access to deeply nested docs without having to resort
        to some recursion to go deep into the result.

        However, if you query for 'analysis', the entire 'analysis' key is
        returned as r['analysis'] and then the subkeys can be accessed in the
        usual form, i.e., r['analysis']['e_above_hull']

        :param properties: Properties to query for. Defaults to None which means all supported properties.
        :param criteria: Criteria to query for as a dict.
        :param distinct_key: If not None, the key for which to get distinct results
        :param \*\*kwargs: Other kwargs supported by pymongo.collection.find.
            Useful examples are limit, skip, sort, etc.
        :return: A QueryResults Iterable, which is somewhat like pymongo's
            cursor except that it performs mapping. In general, the dev does
            not need to concern himself with the form. It is sufficient to know
            that the results are in the form of an iterable of dicts.
        """
        if properties is not None:
            props, prop_dict = self._parse_properties(properties)
        else:
            props, prop_dict = None, None

        crit = self._parse_criteria(criteria)
        if self.query_post:
            for func in self.query_post:
                func(crit, props)
        cur = self.collection.find(filter=crit, projection=props, **kwargs)

        if distinct_key is not None:
            cur = cur.distinct(distinct_key)
            return QueryListResults(prop_dict, cur, postprocess=self.result_post)
        else:
            return QueryResults(prop_dict, cur, postprocess=self.result_post)

    def _parse_properties(self, properties):
        """Make list of properties into 2 things:
        (1) dictionary of { 'aliased-field': 1, ... } for a mongodb query eg. {''}
        (2) dictionary, keyed by aliased field, for display
        """
        props = {}
        # TODO: clean up prop_dict?
        prop_dict = OrderedDict()
        # We use a dict instead of list to provide for a richer syntax
        for p in properties:
            if p in self.aliases:
                if isinstance(properties, dict):
                    props[self.aliases[p]] = properties[p]
                else:
                    props[self.aliases[p]] = 1
                prop_dict[p] = self.aliases[p].split(".")
            else:
                if isinstance(properties, dict):
                    props[p] = properties[p]
                else:
                    props[p] = 1
                prop_dict[p] = p.split(".")
        # including a lower-level key after a higher level key e.g.:
        # {'output': 1, 'output.crystal': 1} instead of
        # {'output.crystal': 1, 'output': 1}
        # causes mongo to skip the other higher level keys.
        # this is a (sketchy) workaround for that. Note this problem
        # doesn't appear often in python2 because the dictionary ordering
        # is more stable.
        props = OrderedDict(sorted(props.items(), reverse=True))
        return props, prop_dict

    def query_one(self, *args, **kwargs):
        """Return first document from :meth:`query`, with same parameters.
        """
        for r in self.query(*args, **kwargs):
            return r
        return None

    def get_structure_from_id(self, task_id, final_structure=True):
        """
        Returns a structure from the database given the task id.

        Args:
            task_id:
                The task_id to query for.
            final_structure:
                Whether to obtain the final or initial structure. Defaults to
                True.
        """
        args = {'task_id': task_id}
        field = 'output.crystal' if final_structure else 'input.crystal'
        results = tuple(self.query([field], args))

        if len(results) > 1:
            raise QueryError("More than one result found for task_id {}!".format(task_id))
        elif len(results) == 0:
            raise QueryError("No structure found for task_id {}!".format(task_id))
        c = results[0]
        return Structure.from_dict(c[field])

    def __repr__(self):
        return "QueryEngine: {}:{}/{}".format(self.host, self.port,
                                              self.database_name)

    @staticmethod
    def from_config(config_file, use_admin=False):
        """
        Initialize a QueryEngine from a JSON config file generated using mgdb
        init.

        Args:
            config_file:
                Filename of config file.
            use_admin:
                If True, the admin user and password in the config file is
                used. Otherwise, the readonly_user and password is used.
                Defaults to False.

        Returns:
            QueryEngine
        """
        with open(config_file) as f:
            d = json.load(f)
            user = d["admin_user"] if use_admin else d["readonly_user"]
            password = d["admin_password"] if use_admin \
                else d["readonly_password"]
            return QueryEngine(
                host=d["host"], port=d["port"], database=d["database"],
                user=user, password=password, collection=d["collection"],
                aliases_config=d.get("aliases_config", None))

    def __getitem__(self, item):
        """Support pymongo.Database syntax db['collection'] to access collections.
        Simply delegate this to the pymongo.Database instance, so behavior is the same.
        """
        return self.db[item]

    def get_dos_from_id(self, task_id):
        """
        Overrides the get_dos_from_id for the MIT gridfs format.
        """
        args = {'task_id': task_id}
        fields = ['calculations']
        structure = self.get_structure_from_id(task_id)
        dosid = None
        for r in self.query(fields, args):
            dosid = r['calculations'][-1]['dos_fs_id']
        if dosid is not None:
            self._fs = gridfs.GridFS(self.db, 'dos_fs')
            with self._fs.get(dosid) as dosfile:
                s = dosfile.read()
                try:
                    d = json.loads(s)
                except:
                    s = zlib.decompress(s)
                    d = json.loads(s.decode("utf-8"))
                tdos = Dos.from_dict(d)
                pdoss = {}
                for i in range(len(d['pdos'])):
                    ados = d['pdos'][i]
                    all_ados = {}
                    for j in range(len(ados)):
                        orb = Orbital(j)
                        odos = ados[str(orb)]
                        all_ados[orb] = {Spin(int(k)): v
                                         for k, v
                                         in odos['densities'].items()}
                    pdoss[structure[i]] = all_ados
                return CompleteDos(structure, tdos, pdoss)
        return None


class QueryResults(Iterable):
    """
    Iterable wrapper for results from QueryEngine.
    Like pymongo's cursor, this object should generally not be instantiated,
    but should be obtained from a queryengine.
    It delegates many attributes to the underlying pymongo cursor, and should
    support nearly all cursor like attributes such as count(), explain(),
    hint(), etc. Please see pymongo cursor documentation for details.
    """
    def __init__(self, prop_dict, result_cursor, postprocess=None):
        """Constructor.

        :param prop_dict: Properties
        :param result_cursor: Iterable returning records
        :param postprocess: List of functions, each taking a record and
            modifying it in-place, or None, or an empty list
        """
        self._results = result_cursor
        self._prop_dict = prop_dict
        self._pproc = postprocess or []  # make empty values iterable


    def _wrapper(self, func):
        """
        This function wraps all callable objects returned by self.__getattr__.
        If the result is a cursor, wrap it into a QueryResults object
        so that you can invoke postprocess functions in self._pproc
        """
        def wrapped(*args, **kwargs):
            ret_val = func(*args, **kwargs)
            if isinstance(ret_val, pymongo.cursor.Cursor):
                ret_val = self.from_cursor(ret_val)
            return ret_val

        return wrapped

    def __getattr__(self, attr):
        """
        Override getattr to make QueryResults inherit all pymongo cursor
        attributes.
        Wrap any callable object with _wrapper to intercept cursors and wrap them
        as a QueryResults object.
        """
        if hasattr(self._results, attr):
            ret_val = getattr(self._results, attr)
            # wrap callable objects to convert returned cursors into QueryResults
            if callable(ret_val):
                return self._wrapper(ret_val)
            return ret_val

    def clone(self):
        return QueryResults(self._prop_dict, self._results.clone())

    def from_cursor(self, cursor):
        """
        Create a QueryResults object from a cursor object
        """
        return QueryResults(self._prop_dict, cursor, self._pproc)


    def __len__(self):
        """Return length as a `count()` on the MongoDB cursor.
        """
        return self._results.count()

    def __getitem__(self, i):
        return self._mapped_result(self._results[i])

    def __iter__(self):
        return self._result_generator()

    def _mapped_result(self, r):
        """Transform/map a result.
        """
        # Apply result_post funcs for pulling out sandbox properties
        for func in self._pproc:
            func(r)
        # If we haven't asked for specific properties, just return object
        if not self._prop_dict:
            result = r
        else:
            result = dict()
            # Map aliased keys back to original key
            for k, v in self._prop_dict.items():
                try:
                    result[k] = self._mapped_result_path(v[1:], data=r[v[0]])
                except (IndexError, KeyError, ValueError):
                    result[k] = None
        return result

    @staticmethod
    def _mapped_result_path(path, data=None):
        if not path:
            return data
        if isinstance(data, list):
            return [QueryResults._mapped_result_path(path, d) for d in data]
        else:
            try:
                return QueryResults._mapped_result_path(path[1:], data[path[0]])
            except (IndexError, KeyError, ValueError):
                return None

    def _result_generator(self):
        for r in self._results:
            yield self._mapped_result(r)


class QueryListResults(QueryResults):
    """Set of QueryResults on a list instead of a MongoDB cursor.
    """
    def clone(self):
        return QueryResults(self._prop_dict, self._results[:])

    def __len__(self):
        """Return length of iterable, as a list if possible; otherwise,
        fall back to the superclass' implementation.
        """
        if hasattr(self._results, '__len__'):
            return len(self._results)
        else:
            return QueryResults.__len__(self)


class QueryError(Exception):
    """
    Exception class for errors occuring during queries.
    """
    pass



"""
Component-level tests for dbgroup module.
"""
__author__ = 'Dan Gunter <dkgunter@lbl.gov>'
__date__ = '5/2/14'

# Stdlib
import json
import logging
import pymongo
import os; _opj = os.path.join
import sys
import tempfile
import time
import unittest
# Package
from matgendb.tests.common import ComponentTest
from matgendb.dbconfig import DB_KEY, COLL_KEY
from matgendb.dbgroup import ConfigGroup

_log = logging.getLogger("comp_dbgroup")
_h = logging.StreamHandler(sys.stdout)
_log.addHandler(_h)
_log.setLevel(logging.INFO)

class DBGroupComponentTest(ComponentTest):
    DB_COLL = {
        "marvel": ["spiderman", "spiderman.amazing", "spiderman.spectacular", "hulk"],
        "dc": ["flash", "flash.garrick", "flash.barry", "flash.wally", "superman"]
    }
    DB_COLL_CFG = {
        "marvel": ["spiderman", "hulk"],
        "dc": ["flash", "superman"]
    }

    def setUp(self):
        self._tmpdir = tempfile.mkdtemp()
        self._config_names = []
        self._conn = pymongo.MongoClient()
        self.create_configs()
        self.create_dbs()

    def tearDown(self):
        for fname in os.listdir(self._tmpdir):
            os.unlink(_opj(self._tmpdir, fname))
        os.rmdir(self._tmpdir)

    def create_configs(self):
        p = self._tmpdir
        cfg = {"host": "localhost", "port": 27017}
        for db, collections in self.DB_COLL_CFG.items():
            cfg[DB_KEY] = db
            for coll in collections:
                cfg[COLL_KEY] = coll
                with open(_opj(p, "{}_{}.json".format(db, coll)), "w") as f:
                    json.dump(cfg, f)
                self._config_names.append("{}.{}".format(db, coll))

    def create_dbs(self):
        for dbname, collections in self.DB_COLL.items():
            db = self._conn[dbname]
            for collname in collections:
                coll = db[collname]
                coll.insert({"hello": collname})

    def test_readall(self):
        """Read all configurations in a directory.
        """
        g = ConfigGroup()
        t0 = time.time()
        g.add_path(self._tmpdir)
        t1 = time.time()
        _log.debug("Time to scan directory = {:.3g}s".format(t1 - t0))
        expect = set(self._config_names)
        got = set(g.keys())
        self.assertEqual(expect, got)

    def test_expand(self):
        """Expand configurations to get full list of collections.
        """
        g = ConfigGroup().add_path(self._tmpdir)
        _log.debug("Base: {}".format(g.keys()))
        t0 = time.time()
        g.expand("marvel.*")
        t1 = time.time()
        _log.debug("Expanded in {:.3g}s: {}".format(t1 - t0, g.keys()))
        # expect expanded marvel, but just configured dc
        marvel = map(lambda val: 'marvel.' + val, self.DB_COLL['marvel'])
        dc = map(lambda val: 'dc.' + val, self.DB_COLL_CFG['dc'])
        expect = set(marvel + dc)
        got = set(g.keys())
        self.assertEqual(expect, got)

    def test_sandbox(self):
        """Combine expand and prefix to make a sandbox.
        """
        g = ConfigGroup().add_path(self._tmpdir)
        for i, sandbox in enumerate(["marvel.spiderman", "dc.flash"]):
            g.expand("{}*".format(sandbox))
            _log.debug("Expanded {} keys: {}".format(sandbox,g.keys()))
            g.set_prefix(sandbox)
            qe, expect_name = None, None
            if i == 0:
                for wow in "amazing", "spectacular":
                    qe = g[wow]
                    expect_name = "spiderman.{}".format(wow)
                    self.assertEqual(qe.collection.name, expect_name)
                    self.assertEqual(qe.db.name, "marvel")
            elif i == 1:
                for who in "garrick", "barry", "wally":
                    qe = g[who]
                    expect_name = "flash.{}".format(who)
                    self.assertEqual(qe.collection.name, expect_name)
                    self.assertEqual(qe.db.name, "dc")
            # only entry in db should be hello:<collname>
            _log.debug("Query QE={} collection={}".format(qe, qe.collection.name))
            cur = qe.query()
            row = cur[0]
            self.assertEqual(row["hello"], expect_name)


if __name__ == '__main__':
    unittest.main()



"""
Run derived collection builder.

See the -h option for details on the options.

Subcommands:

* merge - Merge two collections, e.g., a "sandbox" and core DB.
* list - List builder commands.
* build - Run a builder command.

"""
__author__ = 'Dan Gunter <dkgunter@lbl.gov>'
__date__ = '5/22/13'

# Imports
# -------

# System imports.
import argparse
import importlib
import itertools as itt
import json
import logging
import os
import sys
import time
import traceback

# Third-party imports.
import pymongo

# Local imports.
from matgendb.builders import core
from matgendb.util import csv_list, kvp_dict
from matgendb.query_engine import QueryEngine
from matgendb.builders.incr import TrackedQueryEngine, UnTrackedQueryEngine, Operation

# Global variables.

_log = None     # configured in main()

DEFAULT_CONFIG_FILE = "db.json"

# Suffix for merged tasks collection
MERGED_SUFFIX = "merged"

# Cmdline flag for parameters
NOINCR_FLAG = '-'

# Exceptions
# ----------


class BuilderError(Exception):
    pass


class ConfigurationError(BuilderError):
    def __init__(self, where, why):
        Exception.__init__(self, "class {}: {}".format(where, why))


class BuilderNotFoundError(Exception):
    pass

# Commands
# --------


def command_merge(args):
    """Command: Merge tasks from two collections.
    Intended for merging a sandbox and core db.

    :param args: Command-line arguments
    :type args: list
    """
    # Check args.
    if not args.sandbox_file:
        raise ConfigurationError("sandbox filename",
                                 "In sandbox mode, -s/--sandbox is required")

    # Connect to "core" collections.
    try:
        settings = get_settings(args.config_file)
    except ConfigurationError as err:
        _log.error("command_merge.configuration")
        raise
    core_db = QueryEngine(**settings)
    if hasattr(args, 'merged_tasks') and args.merged_tasks:
        suffix = MERGED_SUFFIX
    else:
        suffix = None
    core_collections = core.Collections(core_db, task_suffix=suffix)

    # Setup.
    sandbox_settings = get_settings(args.sandbox_file)
    sandbox_db = QueryEngine(**sandbox_settings)
    sdb = sandbox_settings['database']
    pfx = args.coll_prefix
    sandbox_collections = core.Collections(sandbox_db, prefix=pfx)
    # set task id prefix
    if args.sandbox_name:
        id_prefix = args.sandbox_name
    elif pfx:
        id_prefix = pfx
    else:
        id_prefix = "sandbox"
    # set target collection name
    if pfx:
        target = "{}.tasks.{}".format(pfx, MERGED_SUFFIX)
    else:
        target = "tasks.{}".format(MERGED_SUFFIX)

    # Perform the merge.
    _log.debug("sandbox.merge.begin: sandbox={}".format(sdb))
    t0 = time.time()
    try:
        core.merge_tasks(core_collections, sandbox_collections, id_prefix, target, wipe=args.wipe_target)
    except pymongo.errors.DuplicateKeyError as err:
        _log.error("sandbox.merge.end error=merge.duplicate_key msg={}".format(err))
        tell_user("\nDuplicate key error from MongoDB.\nUse -W/--wipe to clear target collection before merge.\n")
        return -1
    _log.debug("sandbox.merge.end: sandbox={} duration_sec={:g}".format(sdb, time.time() - t0))
    tell_user("Merged tasks: db={} collection={}".format(sdb, target))

    return 0

def _show_builder(b, params=None, strm=sys.stderr):
    """Print a formatted version of builder info to the console.
    `b` is a builder class.
    """
    info = dict(desc=(b.__doc__ or "(no description)").strip(),
                name='.'.join((b.__module__, b.__name__)))
    strm.write("\n{name}\n  {desc}\n".format(**info))
    if params is None:
        params = core.parse_fn_docstring(b.get_items)
    strm.write("  Parameters:\n")
    for p, pval in params.iteritems():
        info = dict(name=p, type=pval['type'], desc=pval['desc'])
        strm.write("    {name} ({type}): {desc}\n".format(**info))
    strm.write("\n")

def command_build(args):
    """Command: Run builder, invoked from 'build' sub-command

    :param args: Command-line arguments
    :type args: list
    """
    # Get builder class
    builder_class = find_builder(args.builder, args.cls)
    if builder_class is None:
        raise BuilderNotFoundError("{}.{}".format(
            args.builder, args.cls))
    # Create builder
    kwargs = dict(ncores=args.num_cores)
    try:
        builder = builder_class(**kwargs)
    except TypeError as err:
        raise BuilderError("Cannot create object: {}.{}({}): {}".format(
            args.builder, args.cls, kvp_dict(kwargs), err))
    # Get parameter types
    params = builder.get_parameters()
    if params is None:
        # Parse builder's get_items() method docstring.
        _log.debug("parse builder docstring")
        params = core.parse_fn_docstring(builder_class.get_items)

    # Check args against the builder parameters
    if args.usage:
        _show_builder(builder_class, params=params)
        return 0
    user_params, meth_params = [x[0] for x in args.params], params.keys()
    if set(user_params) != set(meth_params):
        if len(user_params) != len(meth_params):
            adj, ineq = (("few", "<"), ("many", ">"))[
                len(user_params) > len(meth_params)]
            msg = "too {} arguments (got {:d} {} wanted {:d})".format(
                adj, len(user_params), ineq, len(meth_params))
        else:
            bad_params = [p for p in user_params if not p in meth_params]
            msg = "unknown argument(s): {}".format(csv_list(bad_params))
        _show_builder(builder_class, params=params)
        raise ConfigurationError(args.builder, msg)

    # Fill builder args using builder parameter types
    simple_types = {k: eval(k) for k in ('int', 'float')}
    eval_types = {k: eval(k) for k in ('dict', 'list')}
    parsed_builder_args, query_engines = {}, []
    for key, value in args.params:
        _type = params[key]['type']
        if _type.endswith("QueryEngine"):
            # Replace path to query engine config with a QE instance
            try:
                qe = create_qe(value, args.incr)
                parsed_builder_args[key] = qe
                query_engines.append(qe)
            except ValueError as err:
                raise ConfigurationError("creating query engine", err)
        elif _type in simple_types:
            parsed_builder_args[key] = simple_types[_type](value)
        elif _type in eval_types:
            parsed_builder_args[key] = eval(value)
        else:
            parsed_builder_args[key] = value

    # Run builder.
    _log.info("run: {}({})".format(builder_class.__name__,
                                   kvp_dict(parsed_builder_args)))
    count = builder.run(user_kw=parsed_builder_args)
    if count < 1:
        _log.warn("Processed {:d} items".format(count))
    else:
        _log.info("Processed {:d} items".format(count))
    # Save current position, for all query engines
    [qe.set_mark() for qe in query_engines]
    result = 0

    return result

# Utility functions
# -----------------

def create_qe(path, incr=None):
    """Configure a new QueryEngine from a config file at `path`.
    """
    if incr is not None and path.startswith(NOINCR_FLAG):
        incr = None
        path = path[1:]

    db_settings = get_settings(path)

    # Pick QueryEngine subclass
    if incr is not None:
        _log.info("Incremental build: {}".format(path))
        iop, ifield = incr['op'], incr['field']  # set by incr_option()
        qe_class = TrackedQueryEngine
        # Update settings to include the type of operation, and field, used
        # to track the position in the source collection.
        try:
            t_op = Operation[iop]
        except KeyError:
            raise ValueError("Bad operation for incremental mode: '{}'"
                             .format(iop))
        db_settings.update(dict(track_operation=t_op, track_field=ifield))
    else:
        _log.info("Non-incremental build: {}".format(path))
        qe_class = UnTrackedQueryEngine

    # Instantiate QueryEngine
    _log.debug("QueryEngine.create settings={}".format(db_settings))
    try:
        qe = qe_class(**db_settings)
    except pymongo.errors.ConnectionFailure as err:
        raise ValueError("Cannot connect from settings in '{}': {}"
                         .format(path, err))
    return qe

def tell_user(message):
    """Print something to the user.
    """
    print(message)


def get_settings(config_file, allow_default=True):
    """Read settings from a configuration file.
    """
    try:
        if config_file:
            cfg = json.load(open(config_file))
        elif allow_default and os.path.exists(DEFAULT_CONFIG_FILE):
            cfg = json.load(open(DEFAULT_CONFIG_FILE))
        else:
            raise ValueError("Default configuration '{}' not found".format(DEFAULT_CONFIG_FILE))
    except Exception as err:
        raise ConfigurationError(config_file, err)
    normalize_userpass(cfg)
    normalize_aliases(cfg)
    return cfg

def find_builder(module_name, class_name):
    """Find a builder class.

    :param module_name: Name of module (dotted)
    :param class_name: Name of builder class.
    :returnss: class obj, or None if not found.
    """
    try:
        md = importlib.import_module(module_name)
    except ImportError:
        _log.info("Failed to import module '{}'".format(module_name))
        return None
    if not class_name in dir(md):
        _log.info("Failed to find class '{}' in module '{}'"
                  .format(class_name, module_name))
        return None
    return getattr(md, class_name)


def normalize_userpass(cfg):
    """In DB conn. config, normalize user/password from readonly and admin prefixes.
    In the end, there will be only keys 'user' and 'password'.
    """
    for pfx in 'readonly', 'admin':  # in reverse order of priority, to overwrite
        if (pfx + '_user') in cfg and (pfx + '_password') in cfg:
            cfg[QueryEngine.USER_KEY] = cfg[pfx + '_user']
            cfg[QueryEngine.PASSWORD_KEY] = cfg[pfx + '_password']
            del cfg[pfx + '_user']
            del cfg[pfx + '_password']


def normalize_aliases(cfg):
    """Normalize the 'aliases_config' used by the QueryEngine class.
    If the user sets something, leave it alone.
    But if there is nothing, use the name of the collection to decide whether
    to let the QueryEngine use its defaults or empty it out.

    :param cfg: Configuration read from JSON
    :type cfg: dict
    :return: None (cfg is modified in-place)
    """
    if QueryEngine.ALIASES_CONFIG_KEY in cfg:
        return  # explicitly set by user, so do nothing
    if cfg[QueryEngine.COLLECTION_KEY] in ('materials', 'tasks'):
        pass  # let the defaults for QueryEngine proceed
    else:
        # explicitly put empty aliases, to override QueryEngine defaults
        cfg[QueryEngine.ALIASES_CONFIG_KEY] = {'aliases': {}, 'defaults': {}}

def kvp_option(arg):
    """Parse key=value pair into (key, value)
    """
    try:
        key, value = arg.split('=', 1)
    except ValueError:
        raise ConfigurationError("Bad key=value pair: {}".format(arg))
    return key, value

def incr_option(arg, default_field='_id'):
    """Parse incremental option of OPERATION[:FIELD]
    into a little dictionary with keys 'op' and 'field'.
    """
    d = {}
    if ':' in arg:
        d['op'], d['field'] = arg.split(':', 1)
    else:
        d['op'], d['field'] = arg, default_field
    return d

# Main program
# ------------


def main():
    global _log


    # Configure parent parser for shared args.
    parent_parser = argparse.ArgumentParser(add_help=False)
    parent_parser.add_argument('--quiet', '-q', dest='quiet', action="store_true", default=False,
                               help="Minimal verbosity.")
    parent_parser.add_argument('--verbose', '-v', dest='vb', action="count", default=0,
                               help="Print more verbose messages to standard error. Repeatable. (default=ERROR)")
    # Set up argument parsing.
    p = argparse.ArgumentParser(description="Build databases")
    subparsers = p.add_subparsers(description="Actions")

    # Merge action.
    subp = subparsers.add_parser("merge", help="Merge sandbox and core database",
                                 parents=[parent_parser])
    subp.set_defaults(func=command_merge)
    subp.add_argument("-c", "--config", dest="config_file", type=str, metavar='FILE', default="db.json",
                      help="Configure database connection from FILE (%(default)s)")
    subp.add_argument("-n", "--name", dest="sandbox_name", type=str, metavar="NAME", default=None,
                      help="Sandbox name, for prefixing merged task IDs. "
                           "If not given, try to use -p/--prefix, then default (sandbox)")
    subp.add_argument("-p", "--prefix", dest="coll_prefix", type=str, metavar='PREFIX', default=None,
                      help="Collection name prefix for input (and possibly output) collections")
    subp.add_argument("-s", "--sandbox", dest="sandbox_file", type=str, metavar='FILE', default=None,
                      help="Configure sandbox database from FILE (required)")
    subp.add_argument("-W", "--wipe", dest="wipe_target", action="store_true",
                      help="Wipe target collection, removing all data in it, before merge")


    # Build actions: called either 'run' or 'build'.
    for action in "build", "run":
        subp = subparsers.add_parser(action, help="run a builder",
                                     parents=[parent_parser])
        subp.set_defaults(func=command_build)
        subp.add_argument("builder", type=str, metavar="builder", default=None,
                          help="Builder module, which is found using "
                               "the module path. e.g., 'my.builders'")
        subp.add_argument("params", nargs="*", metavar="parameter",
                          type=kvp_option,
                          help="Builder parameters, in format <name>=<value>. "
                               "If the parameter type is QueryEngine, the value"
                               " should be a JSON configuration file. "
                               "Prefix filename with a '{}' to ignore "
                               "incremental mode for this QueryEngine."
                               .format(NOINCR_FLAG))
        iops = csv_list(Operation.__members__.keys())
        subp.add_argument("-c", "--class", dest="cls", metavar="NAME",
                          help="Builder class name (default=%(default)s)",
                          default="Builder")
        subp.add_argument("-i", "--incr", dest="incr", type=incr_option,
                          metavar="OPER[:FIELD]",
                          help="Incremental mode for operation and optional "
                               "sort-field name. OPER may be one of: "
                               "{}. Default FIELD is '_id'".format(iops))
        subp.add_argument("-n", "--ncores", dest="num_cores", type=int, default=1,
                          help="Number of cores or processes to run "
                               "in parallel (%(default)d)")
        subp.add_argument("-u", "--usage", action="store_true", dest="usage",
                           help="Print usage information on selected builder "
                                "and exit.")
    # Parse arguments.
    args = p.parse_args()

    # Configure logging.
    _log = logging.getLogger("mg")  # parent
    _log.propagate = False
    hndlr = logging.StreamHandler()
    hndlr.setFormatter(logging.Formatter(
        "[%(levelname)-6s] %(asctime)s %(name)s :: %(message)s"))
    _log.addHandler(hndlr)
    if args.quiet:
        lvl = logging.CRITICAL
    else:
        if args.vb > 1:
            lvl = logging.DEBUG
        elif args.vb > 0:
            lvl = logging.INFO
        else:
            lvl = logging.WARN
    _log.setLevel(lvl)
    _log = logging.getLogger("mg.build")
    # don't send logs up

    # Run function.
    if args.func is None:
        p.error("No action given")
    try:
        result = args.func(args)
    except ConfigurationError as err:
        _log.error("Configuration error: {}".format(err))
        result = -1
    except BuilderNotFoundError as err:
        _log.error("Builder not found: {}".format(err))
        result = -1
    except BuilderError as err:
        _log.error("{}".format(err))
        result = -1
    except Exception as err:
        if _log.getEffectiveLevel() <= logging.INFO:
            print("{}".format(traceback.format_exc()))
        p.error("{} error: {}".format(args.func.__name__, err))
        result = -2
    if result < 0:
        _log.error("Failure: {:d}".format(result))
    return result

if __name__ == '__main__':
    sys.exit(main())



"""
Example builders that read or write files.
"""
__author__ = 'Dan Gunter <dkgunter@lbl.gov>'
__date__ = '5/16/14'

from matgendb.builders.core import Builder

class FileCounter(Builder):
    """Count lines and characters in a file.
    """
    def __init__(self, **kwargs):
        self.num_lines, self.num_chars = 0, 0
        # The flag sequential=True is necessary to
        # allow for the shared variables above. Otherwise,
        # due to parallelism, special procedures would be
        # needed to safely increment shared vars.
        Builder.__init__(self, **kwargs)

    def get_parameters(self):
        return {'input_file': {'type': 'str', 'desc': 'Input file path'}}

    def get_items(self, input_file=None):
        with open(input_file, "r") as f:
            #print("Reading from {}".format(f.name))
            for line in f:
                yield line

    def process_item(self, item):
        self.num_chars += len(item)
        self.num_lines += 1
        #print("{:d} lines, {:d} characters".format(
        #    self.num_lines, self.num_chars))

    def finalize(self, errors):
        print("{:d} lines, {:d} characters".format(
            self.num_lines, self.num_chars))
        return True


"""
Tests for matgendb.util
"""
__author__ = 'dang'

import unittest
from matgendb.util import collection_keys

class FakeCollection:
    """Collection that emulates find_one().
    """
    def __init__(self, data):
        self.data = data

    def find_one(self):
        return self.data


class KeysTestCase(unittest.TestCase):
    """Test collection_keys.
    """
    def test_flat(self):
        keys = ["a", "b", "C"]
        coll = FakeCollection(dict.fromkeys(keys))
        result = collection_keys(coll)
        self.assertEqual(set(keys), set(result))

    def test_nested(self):
        input = {"a": {"b": {"c": 1}}}
        coll = FakeCollection(input)
        result = collection_keys(coll)
        print("@@Result: {}".format(result))
        keys = ["a", "a.b", "a.b.c"]
        self.assertEqual(set(keys), set(result))

if __name__ == '__main__':
    unittest.main()



"""
Collection validator
"""
__author__ = "Dan Gunter"
__copyright__ = "Copyright 2012-2013, The Materials Project"
__version__ = "1.0"
__maintainer__ = "Dan Gunter"
__email__ = "dkgunter@lbl.gov"
__status__ = "Development"
__date__ = "1/31/13"

import pymongo
import random
import sys
import collections

from .util import DoesLogging, total_size
from smoqe.query import *


class DBError(Exception):
    pass

class ValidatorSyntaxError(Exception):
    "Syntax error in configuration of Validator"
    def __init__(self, target, desc):
        msg = 'Invalid syntax: {} -> "{}"'.format(desc, target)
        Exception.__init__(self, msg)

class PythonMethod:
    """Encapsulate an external Python method that will be run on our target
    MongoDB collection to perform arbitrary types of validation.
    """
    _PATTERN = re.compile(r'\s*(@\w+)(\s+\w+)*')

    CANNOT_COMBINE_ERR = 'Call to a Python method cannot be combined '
    'with any other constraints'
    BAD_CONSTRAINT_ERR = 'Invalid constraint (must be: @<method> [<param> ..])'

    @classmethod
    def constraint_is_method(cls, text):
        """Check from the text of the constraint whether it is
        a Python method, as opposed to a 'normal' constraint.

        :return: True if it is, False if not
        """
        m = cls._PATTERN.match(text)
        return m is not None

    def __init__(self, text):
        """Create new instance from a raw constraint string.

        :raises: ValidatorSyntaxerror
        """
        if not self._PATTERN.match(text):
            raise ValidatorSyntaxError(text, self.BAD_CONSTRAINT_ERR)
        tokens = re.split('@?\s+', text)
        if len(tokens) < 1:
            raise ValidatorSyntaxError(text, self.BAD_CONSTRAINT_ERR)
        self.method = tokens[0]
        self.params = tokens[1:]

def mongo_get(rec, key, default=None):
    """
    Get value from dict using MongoDB dot-separated path semantics.
    For example:

    >>> assert mongo_get({'a': {'b': 1}, 'x': 2}, 'a.b') == 1
    >>> assert mongo_get({'a': {'b': 1}, 'x': 2}, 'x') == 2
    >>> assert mongo_get({'a': {'b': 1}, 'x': 2}, 'a.b.c') is None

    :param rec: mongodb document
    :param key: path to mongo value
    :param default: default to return if not found
    :return: value, potentially nested, or default if not found
    :raise: ValueError, if record is not a dict.
    """
    if not rec:
        return default
    if not isinstance(rec, collections.Mapping):
        raise ValueError('input record must act like a dict')
    if not '.' in key:
        return rec.get(key, default)
    for key_part in key.split('.'):
        if not isinstance(rec, collections.Mapping):
            return default
        if not key_part in rec:
            return default
        rec = rec[key_part]
    return rec


class Projection:
    """Fields on which to project the query results.
    """

    def __init__(self):
        self._fields = {}
        self._slices = {}

    def add(self, field, op=None, val=None):
        """Update report fields to include new one, if it doesn't already.

        :param field: The field to include
        :type field: Field
        :param op: Operation
        :type op: ConstraintOperator
        :return: None
        """
        if field.has_subfield():
            self._fields[field.full_name] = 1
        else:
            self._fields[field.name] = 1
        if op and op.is_size() and not op.is_variable():
            # get minimal part of array with slicing,
            # but cannot use slice with variables
            self._slices[field.name] = val + 1
        if op and op.is_variable():
            # add the variable too
            self._fields[val] = 1

    def to_mongo(self):
        """Translate projection to MongoDB query form.

        :return: Dictionary to put into a MongoDB JSON query
        :rtype: dict
        """
        d = copy.copy(self._fields)
        for k, v in self._slices.items():
            d[k] = {'$slice': v}
        return d


class ConstraintViolation:
    """A single constraint violation, with no metadata.
    """
    def __init__(self, constraint, value, expected):
        """Create new constraint violation

        :param constraint: The constraint that was violated
        :type constraint: Constraint
        """
        self._constraint = constraint
        self._got = value
        self._expected = expected

    @property
    def field(self):
        return self._constraint.field.name

    @property
    def op(self):
        #return str(self._constraint.op)
        return self._constraint.op.display_op

    @property
    def got_value(self):
        return self._got

    @property
    def expected_value(self):
        return self._expected

    @expected_value.setter
    def expected_value(self, value):
        self._expected = value


class NullConstraintViolation(ConstraintViolation):
    """Empty constraint violation, for when there are no constraints.
    """
    def __init__(self):
        ConstraintViolation.__init__(self, Constraint('NA', '=', 'NA'), 'NA', 'NA')


class ConstraintViolationGroup:
    """A group of constraint violations with metadata.
    """
    def __init__(self):
        """Create an empty object.
        """
        self._viol = []
        # These are read/write
        self.subject = ''
        self.condition = None

    def add_violations(self, violations, record=None):
        """Add constraint violations and associated record.

        :param violations: List of violations
        :type violations: list(ConstraintViolation)
        :param record: Associated record
        :type record: dict
        :rtype: None
        """
        rec = {} if record is None else record
        for v in violations:
            self._viol.append((v, rec))

    def __iter__(self):
        return iter(self._viol)

    def __len__(self):
        return len(self._viol)


class ProgressMeter:
    """Simple progress tracker
    """
    def __init__(self, num, fmt):
        self._n = num
        self._subject = '?'
        self._fmt = fmt
        self._count = 0
        self._total = 0

    @property
    def count(self):
        return self._total

    def set_subject(self, subj):
        self._subject = subj

    def update(self, *args):
        self._count += 1
        self._total += 1
        if self._n == 0 or self._count < self._n:
            return
        sys.stderr.write(self._fmt.format(*args, subject=self._subject, count=self.count))
        sys.stderr.write('\n')
        sys.stderr.flush()
        self._count = 0


class ConstraintSpec(DoesLogging):
    """Specification of a set of constraints for a collection.
    """
    FILTER_SECT = 'filter'
    CONSTRAINT_SECT = 'constraints'
    SAMPLE_SECT = 'sample'

    def __init__(self, spec):
        """Create specification from a configuration.

        :param spec: Configuration for a single collection
        :type spec: dict
        :raise: ValueError if specification is wrong
        """
        DoesLogging.__init__(self, name='mg.ConstraintSpec')
        self._sections, _slist = {}, []
        for item in spec:
            self._log.debug("build constraint from: {}".format(item))
            if isinstance(item, dict):
                self._add_complex_section(item)
            else:
                self._add_simple_section(item)

    def __iter__(self):
        """Return a list of all the sections.

        :rtype: list(ConstraintSpecSection)
        """
        sect = []
        # simple 1-level flatten operation
        for values in self._sections.values():
            for v in values:
                sect.append(v)
        return iter(sect)

    def _add_complex_section(self, item):
        """Add a section that has a filter and set of constraints

        :raise: ValueError if filter or constraints is missing
        """
        # extract filter and constraints
        try:
            fltr = item[self.FILTER_SECT]
        except KeyError:
            raise ValueError("configuration requires '{}'".format(self.FILTER_SECT))
        sample = item.get(self.SAMPLE_SECT, None)
        constraints = item.get(self.CONSTRAINT_SECT, None)

        section = ConstraintSpecSection(fltr, constraints, sample)
        key = section.get_key()
        if key in self._sections:
            self._sections[key].append(section)
        else:
            self._sections[key] = [section]

    def _add_simple_section(self, item):
        self._sections[None] = [ConstraintSpecSection(None, item, None)]


class ConstraintSpecSection:
    def __init__(self, fltr, constraints, sample):
        self._filter, self._constraints, self._sampler = fltr, constraints, sample
        # make condition(s) into a tuple
        if isinstance(fltr, basestring):
            self._key = (fltr,)
        elif fltr is None:
            self._key = None
        else:
            self._key = tuple(fltr)
        # parse sample keywords into class, if present
        if sample:
            self._sampler = Sampler(**sample)

    def get_key(self):
        return self._key

    @property
    def sampler(self):
        return self._sampler

    @property
    def filters(self):
        return self._filter

    @property
    def constraints(self):
        return self._constraints


class Validator(DoesLogging):
    """Validate a collection.
    """

    class SectionParts:
        """Encapsulate the tuple of information for each section of filters, constraints,
         etc. within a collection.
        """
        def __init__(self, cond, body, sampler, report_fields):
            """Create new initialized set of parts.

            :param cond: Condition to filter records
            :type cond: MongoQuery
            :param body: Main set of constraints
            :type body: MongoQuery
            :param sampler: Sampling class if any
            :type sampler: Sampler
            :param report_fields: Fields to report on
            :type report_fields: list
            """
            self.cond, self.body, self.sampler, self.report_fields = \
                cond, body, sampler, report_fields

    def __init__(self, max_violations=50, max_dberrors=10, aliases=None, add_exists=False):
        DoesLogging.__init__(self, name='mg.validator')
        self.set_progress(0)
        self._aliases = aliases if aliases else {}
        self._max_viol = max_violations
        if self._max_viol > 0:
            self._find_kw = {'limit': self._max_viol}
        else:
            self._find_kw = {}
        self._max_dberr = max_dberrors
        self._base_report_fields = {'_id': 1, 'task_id': 1}
        self._add_exists = add_exists

    def set_aliases(self, a):
        """Set aliases.
        """
        self._aliases = a

    def set_progress(self, num):
        """Report progress every `num` bad records.

        :param num: Report interval
        :type num: int
        :return: None
        """
        report_str = 'Progress for {subject}: {count:d} invalid, {:d} db errors, {:d} bytes'
        self._progress = ProgressMeter(num, report_str)

    def num_violations(self):
        if self._progress is None:
            return 0
        return self._progress._count

    def validate(self, coll, constraint_spec, subject='collection'):
        """Validation of  a collection.
        This is a generator that yields ConstraintViolationGroups.

        :param coll: Mongo collection
        :type coll: pymongo.Collection
        :param constraint_spec: Constraint specification
        :type constraint_spec: ConstraintSpec
        :param subject: Name of the thing being validated
        :type subject: str
        :return: Sets of constraint violation, one for each constraint_section
        :rtype: ConstraintViolationGroup
        :raises: ValidatorSyntaxError
        """
        self._spec = constraint_spec
        self._progress.set_subject(subject)
        self._build(constraint_spec)
        for sect_parts in self._sections:
            cvg = self._validate_section(subject, coll, sect_parts)
            if cvg is not None:
                yield cvg

    def _validate_section(self, subject, coll, parts):
        """Validate one section of a spec.

        :param subject: Name of subject
        :type subject: str
        :param coll: The collection to validate
        :type coll: pymongo.Collection
        :param parts: Section parts
        :type parts: Validator.SectionParts
        :return: Group of constraint violations, if any, otherwise None
        :rtype: ConstraintViolationGroup or None
        """
        cvgroup = ConstraintViolationGroup()
        cvgroup.subject = subject

        # If the constraint is an 'import' of code, treat it differently here
        # if self._is_python(parts):
        #    num_found = self._run_python(cvgroup, coll, parts)
        #    return None if num_found == 0 else cvgroup

        query = parts.cond.to_mongo(disjunction=False)
        query.update(parts.body.to_mongo())
        cvgroup.condition = parts.cond.to_mongo(disjunction=False)
        self._log.debug('Query spec: {}'.format(query))
        self._log.debug('Query fields: {}'.format(parts.report_fields))
        # Find records that violate 1 or more constraints
        cursor = coll.find(query, parts.report_fields, **self._find_kw)
        if parts.sampler is not None:
            cursor = parts.sampler.sample(cursor)
        nbytes, num_dberr, num_rec = 0, 0, 0
        while 1:
            try:
                record = next(cursor)
                nbytes += total_size(record)
                num_rec += 1
            except StopIteration:
                self._log.info("collection {}: {:d} records, {:d} bytes, {:d} db-errors"
                               .format(subject, num_rec, nbytes, num_dberr))
                break
            except pymongo.errors.PyMongoError as err:
                num_dberr += 1
                if num_dberr > self._max_dberr > 0:
                    raise DBError("Too many errors")
                self._log.warn("DB.{:d}: {}".format(num_dberr, err))
                continue

            # report progress
            if self._progress:
                self._progress.update(num_dberr, nbytes)
            # get reasons for badness
            violations = self._get_violations(parts.body, record)
            cvgroup.add_violations(violations, record)
        return None if nbytes == 0 else cvgroup

    def _get_violations(self, query, record):
        """Reverse-engineer the query to figure out why a record was selected.

        :param query: MongoDB query
        :type query: MongQuery
        :param record: Record in question
        :type record: dict
        :return: Reasons why bad
        :rtype: list(ConstraintViolation)
        """
        # special case, when no constraints are given
        if len(query.all_clauses) == 0:
            return [NullConstraintViolation()]
        # normal case, check all the constraints
        reasons = []
        for clause in query.all_clauses:
            var_name = None
            key = clause.constraint.field.name
            op = clause.constraint.op
            fval = mongo_get(record, key)
            if fval is None:
                expected = clause.constraint.value
                reasons.append(ConstraintViolation(clause.constraint, 'missing', expected))
                continue
            if op.is_variable():
                # retrieve value for variable
                var_name = clause.constraint.value
                value = mongo_get(record, var_name, default=None)
                if value is None:
                    reasons.append(ConstraintViolation(clause.constraint, 'missing', var_name))
                    continue
                clause.constraint.value = value         # swap out value, temporarily
            # take length for size
            if op.is_size():
                if isinstance(fval, str) or not hasattr(fval, '__len__'):
                    reasons.append(ConstraintViolation(clause.constraint, type(fval), 'sequence'))
                    if op.is_variable():
                        clause.constraint.value = var_name      # put original value back
                    continue
                fval = len(fval)
            ok, expected = clause.constraint.passes(fval)
            if not ok:
                reasons.append(ConstraintViolation(clause.constraint, fval, expected))
            if op.is_variable():
                clause.constraint.value = var_name      # put original value back
        return reasons

    def _build(self, constraint_spec):
        """Generate queries to execute.

        Sets instance variables so that Mongo query strings, etc. can now
        be extracted from the object.

        :param constraint_spec: Constraint specification
        :type constraint_spec: ConstraintSpec
        """
        self._sections = []

        # For each condition in the spec

        for sval in constraint_spec:
            rpt_fld = self._base_report_fields.copy()
            #print("@@ CONDS = {}".format(sval.filters))
            #print("@@ MAIN = {}".format(sval.constraints))

            # Constraints

            # If the constraint is an external call to Python code
            if self._is_python(sval.constraints):
                query, proj = self._process_python(sval.constraints)
                rpt_fld.update(proj.to_mongo())

            # All other constraints, e.g. 'foo > 12'
            else:
                query = MongoQuery()
                if sval.constraints is not None:
                    groups = self._process_constraint_expressions(sval.constraints)
                    projection = Projection()
                    for cg in groups.values():
                        for c in cg:
                            projection.add(c.field, c.op, c.value)
                            query.add_clause(MongoClause(c))
                        if self._add_exists:
                            for c in cg.existence_constraints:
                                query.add_clause(MongoClause(c, exists_main=True))
                    rpt_fld.update(projection.to_mongo())

            # Filters

            cond_query = MongoQuery()
            if sval.filters is not None:
                cond_groups = self._process_constraint_expressions(sval.filters, rev=False)
                for cg in cond_groups.values():
                    for c in cg:
                        cond_query.add_clause(MongoClause(c, rev=False))

            # Done. Add a new 'SectionPart' for the filter and constraint

            result = self.SectionParts(cond_query, query, sval.sampler, rpt_fld)
            self._sections.append(result)

    def _process_constraint_expressions(self, expr_list, conflict_check=True, rev=True):
        """Create and return constraints from expressions in expr_list.

        :param expr_list: The expressions
        :conflict_check: If True, check for conflicting expressions within each field
        :return: Constraints grouped by field (the key is the field name)
        :rtype: dict
        """
        # process expressions, grouping by field
        groups = {}
        for expr in expr_list:
            field, raw_op, val = parse_expr(expr)
            op = ConstraintOperator(raw_op)
            if field not in groups:
                groups[field] = ConstraintGroup(Field(field, self._aliases))
            groups[field].add_constraint(op, val)

        # add existence constraints
        for cgroup in groups.values():
            cgroup.add_existence(rev)

        # optionally check for conflicts
        if conflict_check:
            # check for conflicts in each group
            for field_name, group in groups.items():
                conflicts = group.get_conflicts()
                if conflicts:
                    raise ValueError('Conflicts for field {}: {}'.format(field_name, conflicts))
        return groups

    def _is_python(self, constraint_list):
        """Check whether constraint is an import of Python code.

        :param constraint_list: List of raw constraints from YAML file
        :type constraint_list: list(str)
        :return: True if this refers to an import of code, False otherwise
        :raises: ValidatorSyntaxError
        """
        if len(constraint_list) == 1 and \
                PythonMethod.constraint_is_method(constraint_list[0]):
            return True
        if len(constraint_list) > 1 and \
                any(filter(PythonMethod.constraint_is_method, constraint_list)):
            condensed_list = '/'.join(constraint_list)
            err = PythonMethod.CANNOT_COMBINE_ERR
            raise ValidatorSyntaxError(condensed_list, err)
        return False

    def _process_python(self, expr_list):
        """Create a wrapper for a call to some external Python code.

        :param expr_list: The expressions
        :return: Tuple of (query, field-projection)
        :rtype: (PythonMethod, Projection)
        """
        return None, None

    def set_aliases(self, new_value):
        "Set aliases and wrap errors in ValueError"
        try:
            self.aliases = new_value
        except Exception as err:
            raise ValueError("invalid value: {}".format(err))

class Sampler(DoesLogging):
    """Randomly sample a proportion of the full collection.
    """

    # Random uniform distribution
    DIST_RUNIF = 1
    # Default distribution
    DEFAULT_DIST = DIST_RUNIF
    # Names of distributions
    DIST_CODES = {'uniform': DIST_RUNIF}

    def __init__(self, min_items=0, max_items=1e9, p=1.0, distrib=DEFAULT_DIST, **kw):
        """Create new parameterized sampler.

        :param min_items: Minimum number of items in the sample
        :param max_items: Maximum number of items in the sample
        :param p: Probability of selecting an item
        :param distrib: Probability distribution code, one of DIST_<name> in this class
        :type distrib: str or int
        :raise: ValueError, if `distrib` is an unknown code or string
        """
        DoesLogging.__init__(self, 'mg.sampler')
        # Sanity checks
        if min_items < 0:
            raise ValueError('min_items cannot be negative ({:d})'.format(min_items))
        if (max_items != 0) and (max_items < min_items):
            raise ValueError('max_items must be zero or >= min_items ({:d} < {:d})'.format(max_items, min_items))
        if not (0.0 <= p <= 1.0):
            raise ValueError('probability, p, must be between 0 and 1 ({:f})'.format(p))
        self.min_items = min_items
        self.max_items = max_items
        self.p = p
        self._empty = True
        # Distribution
        if not isinstance(distrib, int):
            distrib = self.DIST_CODES.get(str(distrib), None)
        if distrib == self.DIST_RUNIF:
            self._keep = self._keep_runif
        else:
            raise ValueError("unrecognized distribution: {}".format(distrib))

    @property
    def is_empty(self):
        return self._empty

    def _keep_runif(self):
        return self.p >= random.uniform(0, 1)

    def sample(self, cursor):
        """Extract records randomly from the database.
        Continue until the target proportion of the items have been
        extracted, or until `min_items` if this is larger.
        If `max_items` is non-negative, do not extract more than these.

        This function is a generator, yielding items incrementally.

        :param cursor: Cursor to sample
        :type cursor: pymongo.cursor.Cursor
        :return: yields each item
        :rtype: dict
        :raise: ValueError, if max_items is valid and less than `min_items`
                or if target collection is empty
        """
        count = cursor.count()

        # special case: empty collection
        if count == 0:
            self._empty = True
            raise ValueError("Empty collection")

        # special case: entire collection
        if self.p >= 1 and self.max_items <= 0:
            for item in cursor:
                yield item
            return

        # calculate target number of items to select
        if self.max_items <= 0:
            n_target = max(self.min_items, self.p * count)
        else:
            if self.p <= 0:
                n_target = max(self.min_items, self.max_items)
            else:
                n_target = max(self.min_items, min(self.max_items, self.p * count))
        if n_target == 0:
            raise ValueError("No items requested")

        # select first `n_target` items that pop up with
        # probability self.p
        # This is actually biased to items at the beginning
        # of the file if n_target is smaller than (p * count),
        n = 0
        while n < n_target:
            try:
                item = next(cursor)
            except StopIteration:
                # need to keep looping through data until
                # we get all our items!
                cursor.rewind()
                item = next(cursor)
            if self._keep():
                yield item
                n += 1



#!/usr/bin/env python

"""
TODO: Change the module doc.
"""

from __future__ import division

__author__ = "shyuepingong"
__version__ = "0.1"
__maintainer__ = "Shyue Ping Ong"
__email__ = "shyuep@gmail.com"
__status__ = "Beta"
__date__ = "2/26/13"




"""
Pymatgen-db is a database add-on for the Python Materials Genomics (pymatgen)
materials analysis library. It enables the creation of Materials
Project-style MongoDB databases for management of materials data and
provides a clean and intuitive web ui for exploring that data. A query engine
is also provided to enable the easy translation of MongoDB docs to useful
pymatgen objects for analysis purposes.
"""

import os

__author__ = "Shyue Ping Ong, Dan Gunter"
__date__ = "Jul 22 2017"
__version__ = "2019.3.28"


from .query_engine import QueryEngine

SETTINGS_FILE = os.path.join(os.path.expanduser('~'), '.pmgrc.yaml')


def _load_mgdb_settings():
    try:
        import yaml
        with open(SETTINGS_FILE, 'rt') as f:
            d = yaml.load(f)
    except IOError:
        return {}
    return d

SETTINGS = _load_mgdb_settings()


"""
Common functions for tests
"""
__author__ = 'Dan Gunter <dkgunter@lbl.gov>'
__date__ = '10/29/13'

# Stdlib
import json
import logging
import os
import subprocess
import sys
import tempfile
import traceback
import unittest
# Third-party
from mongomock import MongoClient
import pymongo
# Package
from matgendb.query_engine import QueryEngine
from matgendb.builders.incr import CollectionTracker

_log = logging.getLogger('matgendb.tests')

TEST_FILES_DIR = os.path.join(
    os.path.dirname(os.path.dirname(os.path.dirname(os.path.abspath(__file__)))),
    "test_files")


def has_mongo():
    """Determine if MongoDB is up and usable
    """
    if os.environ.get('MP_FAKEMONGO'):
        mongo = False
    else:
        try:
            pymongo.MongoClient()
            mongo = True
        except:
            mongo = False
    return mongo

class MockQueryEngine(QueryEngine):
    """Mock (fake) QueryEngine, unless a real connection works.
    You can disable the attempt to do a real connection
    by setting MP_FAKEMONGO to anything
    """
    def __init__(self, host="127.0.0.1", port=27017, database="vasp",
                 user=None, password=None, collection="tasks",
                 aliases_config=None, default_properties=None):
        if has_mongo():
            try:
                QueryEngine.__init__(self, host=host, port=port,
                                     database=database,
                                     user=user, password=password,
                                     collection=collection,
                                     aliases_config=aliases_config,
                                     default_properties=default_properties)
                _log.warning("Connected to real MongoDB at {}:{}".format(host, port))
                return  # actully connected! not mocked..
            except:
                _log.debug("Connection to real MongoDB at {}:{} failed. "
                           "This is normal; using mock."
                           .format(host, port))
        self.connection = MongoClient(host, port)
        self.db = self.connection[database]
        self._user, self._password = user, password
        self.host = host
        self.port = port
        self.database_name = database
        # colllection name is now a @property. the setter will set "self.collection" internally
        self.collection_name = collection
        self.set_aliases_and_defaults(aliases_config=aliases_config,
                                      default_properties=default_properties)

# -----------------------------------
# Component test classes / functions
# -----------------------------------

def get_component_logger(name, strm=sys.stdout):
    log = logging.getLogger(name)
    if 'TEST_DEBUG' in os.environ:
        log.setLevel(logging.DEBUG)
    else:
        log.setLevel(logging.INFO)
    _h = logging.StreamHandler(strm)
    log.addHandler(_h)
    return log

class ComponentTest(unittest.TestCase):
    DB = 'testdb'
    SRC = 'source'
    DST = 'dest'

    MGBUILD_CMD = ["mgbuild", "run"]

    def setUp(self):
        self.db = self.connect(True)
        self.src, self.dst = self.db[self.SRC], self.db[self.DST]
        self.src_conf, self.dst_conf = self.create_configs()

    def mgbuild(self, args):
        try:
            s = subprocess.check_output(self.MGBUILD_CMD + args,
                                        stderr=subprocess.STDOUT)
        except subprocess.CalledProcessError as err:
            print("ERROR: {}".format(err.output))
            raise
        return s

    def connect(self, clear=False):
        """Connect to Mongo DB

        :return: pymongo Database
        """
        c = pymongo.MongoClient()
        db = c[self.DB]
        if clear:
            for coll in self.SRC, self.DST:
                db[coll].remove()
                tcoll = coll + '.' + CollectionTracker.TRACKING_NAME
                db[tcoll].remove() # remove tracking as well
        return db

    def get_record(self, i):
        return {
            "number": i,
            "data": [
                1, 2, 3
            ],
            "name": "mp-{:d}".format(i)
        }

    def add_records(self, coll, n):
        for i in range(n):
            coll.insert(self.get_record(i))

    def create_configs(self):
        base = {"host": "localhost",
                "port": 27017,
                "database": self.DB,
                "collection": None}
        files = []
        for coll in (self.SRC, self.DST):
            f = tempfile.NamedTemporaryFile(suffix=".json")
            base['collection'] = coll
            json.dump(base, f)
            f.flush()
            files.append(f)
        return files

    def tearDown(self):
        pass

    def run_command(self, args, options):
        """Run the command-line given by the list
        in `args`, adding the dictionary given by
        options as long-form --{key}=value pairs.
        """
        for key, value in options:
            args.append("--{}".format(key))
            if value:
                args.append(value)
        return subprocess.call(args)
